from xares.task import TaskConfig


def cremad_config(encoder) -> TaskConfig:
    return TaskConfig(
        encoder=encoder,
        label_processor=lambda x: {"H": 0, "S": 1, "A": 2, "F": 3, "D": 4, "N": 5}.get(x["label"], -1),
        name="cremad",
        output_dim=6,
        test_split="wds-audio-test",
        train_split="wds-audio-train",
        valid_split="wds-audio-valid",
        zenodo_id="14646870",
    )



from xares.task import TaskConfig


def fsdkaggle2018_config(encoder) -> TaskConfig:
    data_key = "sound"
    class_label_maps = {
        "Hi-hat": 0,
        "Saxophone": 1,
        "Trumpet": 2,
        "Glockenspiel": 3,
        "Cello": 4,
        "Knock": 5,
        "Gunshot_or_gunfire": 6,
        "Clarinet": 7,
        "Computer_keyboard": 8,
        "Keys_jangling": 9,
        "Snare_drum": 10,
        "Writing": 11,
        "Laughter": 12,
        "Tearing": 13,
        "Fart": 14,
        "Oboe": 15,
        "Flute": 16,
        "Cough": 17,
        "Telephone": 18,
        "Bark": 19,
        "Chime": 20,
        "Bass_drum": 21,
        "Bus": 22,
        "Squeak": 23,
        "Scissors": 24,
        "Harmonica": 25,
        "Gong": 26,
        "Microwave_oven": 27,
        "Burping_or_eructation": 28,
        "Double_bass": 29,
        "Shatter": 30,
        "Fireworks": 31,
        "Tambourine": 32,
        "Cowbell": 33,
        "Electric_piano": 34,
        "Meow": 35,
        "Drawer_open_or_close": 36,
        "Applause": 37,
        "Acoustic_guitar": 38,
        "Violin_or_fiddle": 39,
        "Finger_snapping": 40,
    }
    return TaskConfig(
        batch_size_train=64,
        do_knn=False,
        encoder=encoder,
        epochs=3,
        label_processor=lambda x: class_label_maps[x[data_key]],
        learning_rate=1e-3,
        name="fsdkaggle2018",
        output_dim=len(class_label_maps),
        test_split="fsd18_test",
        train_split="fsd18_train",
        valid_split="fsd18_test",
        zenodo_id="14725117",
    )



from xares.task import TaskConfig


def esc50_config(encoder) -> TaskConfig:
    config = TaskConfig(
        encoder=encoder,
        k_fold_splits=list(range(1, 6)),
        label_processor=lambda x: x["label"],
        name="esc50",
        output_dim=50,
        zenodo_id="14614287",
    )

    config.audio_tar_name_of_split = {fold: f"wds-audio-fold-{fold}-*.tar" for fold in config.k_fold_splits}
    config.encoded_tar_name_of_split = {fold: f"wds-encoded-fold-{fold}-*.tar" for fold in config.k_fold_splits}

    return config



from xares.task import TaskConfig


def fluentspeechcommands_config(encoder) -> TaskConfig:
    data_key = "transcription"
    class_label_maps = {
        "Change language": 0,
        "Resume": 1,
        "Turn the lights on": 2,
        "Switch on the lights": 3,
        "Switch off the lights": 4,
        "Volume up": 5,
        "Turn the volume up": 6,
        "Turn the volume down": 7,
        "Turn up the temperature": 8,
        "Turn the heat up": 9,
        "Turn down the temperature": 10,
        "Turn the temperature down": 11,
        "Turn down the heat": 12,
        "Turn the heat down": 13,
        "Turn off the music": 14,
        "Stop the music": 15,
        "Put on the music": 16,
        "Play the music": 17,
        "Start the music": 18,
        "Turn on the lamp": 19,
        "Switch on the lamp": 20,
        "Lamp on": 21,
        "Turn off the lamp": 22,
        "Turn the lamp off": 23,
        "Lamp off": 24,
        "Turn on the kitchen lights": 25,
        "Switch on the kitchen lights": 26,
        "Turn the lights on in the kitchen": 27,
        "Switch on the lights in the kitchen": 28,
        "Lights on in the kitchen": 29,
        "Kitchen lights on": 30,
        "Switch the kitchen lights on": 31,
        "Turn the kitchen lights on": 32,
        "Turn on the lights in the bedroom": 33,
        "Switch on the lights in the bedroom": 34,
        "Turn on the washroom lights": 35,
        "Turn the washroom lights on": 36,
        "Switch on the washroom lights": 37,
        "Lights on in the washroom": 38,
        "Washroom lights on": 39,
        "Turn on the bathroom lights": 40,
        "Turn the bathroom lights on": 41,
        "Switch on the bathroom lights": 42,
        "Lights on in the bathroom": 43,
        "Bathroom lights on": 44,
        "Switch the lights off": 45,
        "Turn the lights off in the kitchen": 46,
        "Switch off the lights in the kitchen": 47,
        "Switch the lights off in the kitchen": 48,
        "Kitchen lights off": 49,
        "Turn the bedroom lights off": 50,
        "Switch the bedroom lights off": 51,
        "Bedroom lights off": 52,
        "Switch off the washroom lights": 53,
        "Lights off in the washroom": 54,
        "Washroom lights off": 55,
        "Turn it up": 56,
        "Volume max": 57,
        "Louder please": 58,
        "Increase the volume": 59,
        "Increase the sound volume": 60,
        "I need volume": 61,
        "Louder phone": 62,
        "I need to hear this, increase the volume": 63,
        "I couldn't hear anything, turn up the volume": 64,
        "This video sound is too low, turn up the volume": 65,
        "Turn sound down": 66,
        "Decrease audio volume": 67,
        "Reduce audio volume": 68,
        "Quieter": 69,
        "Volume mute": 70,
        "Lower the volume": 71,
        "Decrease sound levels": 72,
        "Too loud": 73,
        "That\u2019s too loud": 74,
        "Far too loud": 75,
        "It\u2019s too loud, turn it down": 76,
        "It\u2019s too loud, turn the volume down": 77,
        "Increase the temperature": 78,
        "Increase the heating": 79,
        "Could you increase the heating please?": 80,
        "Make it hotter": 81,
        "More heat": 82,
        "Heat up": 83,
        "Turn the kitchen temperature up": 84,
        "Increase the temperature in the kitchen": 85,
        "Increase the heating in the kitchen": 86,
        "Turn up the heat in the kitchen": 87,
        "Turn up the temperature in the bedroom": 88,
        "Turn the temperature in the bedroom up": 89,
        "Turn up the bedroom heat": 90,
        "Bedroom heat up": 91,
        "Increase the temperature in the bedroom": 92,
        "Increase the heating in the bedroom": 93,
        "Turn up the washroom temperature": 94,
        "Washroom heat up": 95,
        "Increase the temperature in the washroom": 96,
        "Turn up the bathroom temperature": 97,
        "Turn the temperature up in the bathroom": 98,
        "Turn up the heat in the bathroom": 99,
        "Turn the heat up in the bathroom": 100,
        "Bathroom heat up": 101,
        "Increase the temperature in the bathroom": 102,
        "Increase the heating in the bathroom": 103,
        "Less heat": 104,
        "Heat down": 105,
        "Could you decrease the heating?": 106,
        "Turn the kitchen temperature down": 107,
        "Turn the heat down in the kitchen": 108,
        "Kitchen heat down": 109,
        "Decrease the temperature in the kitchen": 110,
        "Turn down the bedroom heat": 111,
        "Turn the bedroom heat down": 112,
        "Bedroom heat down": 113,
        "Decrease the temperature in the bedroom": 114,
        "Decrease the heating in the bedroom": 115,
        "Turn the temperature down in the washroom": 116,
        "Turn down the heat in the washroom": 117,
        "Decrease the heating in the washroom": 118,
        "Turn down the bathroom temperature": 119,
        "Turn down the heat in the bathroom": 120,
        "Turn the heat down in the bathroom": 121,
        "Decrease the temperature in the bathroom": 122,
        "Decrease the heating in the bathroom": 123,
        "Bathroom heat down": 124,
        "Pause music": 125,
        "Stop": 126,
        "Get me the newspaper": 127,
        "Bring the newspaper": 128,
        "Fetch the newspaper": 129,
        "Bring some juice": 130,
        "Bring socks": 131,
        "Go get me my socks": 132,
        "Fetch my socks": 133,
        "Switch the language": 134,
        "Change system language": 135,
        "Open language settings": 136,
        "Set the language": 137,
        "Language settings": 138,
        "Set language to Chinese": 139,
        "Set language to Korean": 140,
        "Set my device to Chinese": 141,
        "OK now switch the main language to Chinese": 142,
        "Set my phone's language to English": 143,
        "I need to practice my German. Switch the language": 144,
        "I need to practice my English. Switch the language": 145,
        "Turn the washroom lights off": 146,
        "Turn off the lights": 147,
        "Pause the music": 148,
        "Volume down": 149,
        "Lights on": 150,
        "Turn the lights off": 151,
        "Lights off": 152,
        "Make the music softer": 153,
        "Lights on in the bedroom": 154,
        "Turn sound up": 155,
        "Make it louder": 156,
        "Louder": 157,
        "Too quiet": 158,
        "Turn it down": 159,
        "Volume lower": 160,
        "Could you increase the heating?": 161,
        "Turn up the temperature in the kitchen": 162,
        "Turn the heat up in the kitchen": 163,
        "Kitchen heat up": 164,
        "Turn the heat up in the washroom": 165,
        "Increase the heating in the washroom": 166,
        "Make it cooler": 167,
        "Decrease the temperature": 168,
        "Decrease the heating": 169,
        "Turn down the heat in the kitchen": 170,
        "Decrease the heating in the kitchen": 171,
        "Turn down the temperature in the bedroom": 172,
        "Turn the temperature in the bedroom down": 173,
        "Turn the heat down in the washroom": 174,
        "Washroom heat down": 175,
        "Turn the temperature down in the bathroom": 176,
        "Play music": 177,
        "Bring newspaper": 178,
        "Go get me some juice": 179,
        "Bring me my socks": 180,
        "Get me my socks": 181,
        "Bring my socks": 182,
        "Set language to German": 183,
        "Set language to English": 184,
        "OK now switch the main language to Korean": 185,
        "OK now switch the main language to English": 186,
        "Set my phone's language to Chinese": 187,
        "I need to practice my Korean. Switch the language": 188,
        "Turn on the lights": 189,
        "Turn up the volume": 190,
        "Make the music louder": 191,
        "Turn down the volume": 192,
        "Turn the temperature up": 193,
        "Turn up the heat": 194,
        "Play": 195,
        "Turn the lamp on": 196,
        "Switch off the lamp": 197,
        "Switch the lights on": 198,
        "Turn the bedroom lights on": 199,
        "Switch the bedroom lights on": 200,
        "Bedroom lights on": 201,
        "Turn off the kitchen lights": 202,
        "Lights off in the kitchen": 203,
        "Turn off the lights in the bedroom": 204,
        "Switch off the lights in the bedroom": 205,
        "Lights off in the bedroom": 206,
        "Turn off the washroom lights": 207,
        "Turn volume up": 208,
        "Increase the sound": 209,
        "I can't hear that": 210,
        "That's too quiet": 211,
        "Far too quiet": 212,
        "Turn the sound up": 213,
        "Decrease volume": 214,
        "Turn volume down": 215,
        "Decrease the volume": 216,
        "Make it quieter": 217,
        "Turn the bedroom heat up": 218,
        "Turn the temperature up in the washroom": 219,
        "Turn up the heat in the washroom": 220,
        "Could you decrease the heating please?": 221,
        "Turn down the temperature in the kitchen": 222,
        "Turn down the washroom temperature": 223,
        "Decrease the temperature in the washroom": 224,
        "Stop music": 225,
        "Resume music": 226,
        "Bring me the newspaper": 227,
        "Go get the newspaper": 228,
        "Bring me some juice": 229,
        "Get me some juice": 230,
        "Bring juice": 231,
        "Bring me my shoes": 232,
        "Get me my shoes": 233,
        "Bring shoes": 234,
        "Bring my shoes": 235,
        "Go get me my shoes": 236,
        "Fetch my shoes": 237,
        "Allow a different language": 238,
        "Use a different language": 239,
        "Change the language": 240,
        "Switch language": 241,
        "Switch languages": 242,
        "Set my phone's language to German": 243,
        "Set my phone's language to Korean": 244,
        "I need to practice my Chinese. Switch the language": 245,
        "Pause": 246,
        "OK now switch the main language to German": 247,
    }

    return TaskConfig(
        encoder=encoder,
        label_processor=lambda x: class_label_maps[x[data_key]],
        name="fluentspeechcommands",
        output_dim=len(class_label_maps),
        test_split="fluentspeechcommands_test",
        train_split="fluentspeechcommands_train",
        valid_split="fluentspeechcommands_valid",
        zenodo_id="14722453",
    )



from math import ceil

import torch

from xares.task import TaskConfig


def maestro_config(encoder) -> TaskConfig:
    class_label_maps = {
        21: 0,
        22: 1,
        23: 2,
        24: 3,
        25: 4,
        26: 5,
        27: 6,
        28: 7,
        29: 8,
        30: 9,
        31: 10,
        32: 11,
        33: 12,
        34: 13,
        35: 14,
        36: 15,
        37: 16,
        38: 17,
        39: 18,
        40: 19,
        41: 20,
        42: 21,
        43: 22,
        44: 23,
        45: 24,
        46: 25,
        47: 26,
        48: 27,
        49: 28,
        50: 29,
        51: 30,
        52: 31,
        53: 32,
        54: 33,
        55: 34,
        56: 35,
        57: 36,
        58: 37,
        59: 38,
        60: 39,
        61: 40,
        62: 41,
        63: 42,
        64: 43,
        65: 44,
        66: 45,
        67: 46,
        68: 47,
        69: 48,
        70: 49,
        71: 50,
        72: 51,
        73: 52,
        74: 53,
        75: 54,
        76: 55,
        77: 56,
        78: 57,
        79: 58,
        80: 59,
        81: 60,
        82: 61,
        83: 62,
        84: 63,
        85: 64,
        86: 65,
        87: 66,
        88: 67,
        89: 68,
        90: 69,
        91: 70,
        92: 71,
        93: 72,
        94: 73,
        95: 74,
        96: 75,
        97: 76,
        98: 77,
        99: 78,
        100: 79,
        101: 80,
        102: 81,
        103: 82,
        104: 83,
        105: 84,
        106: 85,
        107: 86,
    }

    def sec_to_frames(sec):
        return ceil(sec / (encoder.hop_size_in_ms / 1000))

    def label_processor(sample):
        maximal_length = max(item["end"] for item in sample)
        maximal_length = sec_to_frames(maximal_length)
        target_vector = torch.zeros((len(class_label_maps), maximal_length), dtype=torch.float32)
        for item in sample:
            start, end, note = sec_to_frames(item["start"]), sec_to_frames(item["end"]), item["note"]
            label_idx = class_label_maps[note]
            target_vector[label_idx, start:end] = 1.0
        return target_vector

    config = TaskConfig(
        batch_size_encode=1,  # Long samples
        batch_size_train=1,  # Samples are very long, avoid extreme padding
        criterion="BCEWithLogitsLoss",
        do_knn=False,
        disabled=True,
        encoder=encoder,
        epochs=50,
        label_processor=label_processor,
        metric_args=dict(hop_size_in_ms=encoder.hop_size_in_ms if encoder else 0, segment_length_in_s=0.1),
        metric="segmentf1",
        name="maestro",
        output_dim=len(class_label_maps),
        task_type="frame",
        test_split="maestro_test",
        train_split="maestro_train",
        valid_split="maestro_valid",
        zenodo_id="14858022",
    )
    return config



from xares.task import TaskConfig


def vocalimiations_config(encoder) -> TaskConfig:
    task = "imitation"
    class_label_maps = {
        "000Animal_Domestic animals_ pets_Cat_Growling_reference": 0,
        "001Animal_Domestic animals_ pets_Cat_Hiss_reference": 1,
        "002Animal_Domestic animals_ pets_Cat_Meow_reference": 2,
        "003Animal_Domestic animals_ pets_Cat_Purr_reference": 3,
        "004Animal_Domestic animals_ pets_Dog_Bark_reference": 4,
        "005Animal_Domestic animals_ pets_Dog_Growling_reference": 5,
        "006Animal_Domestic animals_ pets_Dog_Howl_reference": 6,
        "007Animal_Livestock_ farm animals_ working animals_Cattle_ bovinae_Cowbell_reference": 7,
        "008Animal_Livestock_ farm animals_ working animals_Cattle_ bovinae_Moo_reference": 8,
        "009Animal_Livestock_ farm animals_ working animals_Fowl_Chicken_ rooster_Cluck_reference": 9,
        "010Animal_Livestock_ farm animals_ working animals_Fowl_Chicken_ rooster_Crowing_ cock-a-doodle-doo_reference": 10,
        "011Animal_Livestock_ farm animals_ working animals_Fowl_Duck_Quack_reference": 11,
        "012Animal_Livestock_ farm animals_ working animals_Fowl_Goose_Honk_reference": 12,
        "013Animal_Livestock_ farm animals_ working animals_Fowl_Turkey_Gobble_reference": 13,
        "014Animal_Livestock_ farm animals_ working animals_Goat_Bleat_reference": 14,
        "015Animal_Livestock_ farm animals_ working animals_Horse_Clip-clop_reference": 15,
        "016Animal_Livestock_ farm animals_ working animals_Horse_Neigh_ whinny_reference": 16,
        "017Animal_Livestock_ farm animals_ working animals_Horse_Snort (horse)_reference": 17,
        "018Animal_Livestock_ farm animals_ working animals_Pig_Oink_reference": 18,
        "019Animal_Wild animals_Bird_Bird flight_ flapping wings_reference": 19,
        "020Animal_Wild animals_Bird_Bird vocalization_ bird call_ bird song_Chirp_ tweet_reference": 20,
        "021Animal_Wild animals_Bird_Crow_Caw_reference": 21,
        "022Animal_Wild animals_Bird_Owl_Hoot_reference": 22,
        "023Animal_Wild animals_Bird_Pigeon_ dove_Coo_reference": 23,
        "024Animal_Wild animals_Frog_Croak_reference": 24,
        "025Animal_Wild animals_Insect_Bee_ wasp_ etc_Buzz_reference": 25,
        "026Animal_Wild animals_Insect_Cricket_reference": 26,
        "027Animal_Wild animals_Insect_Mosquito_reference": 27,
        "028Animal_Wild animals_Roaring cats (lions_ tigers)_Roar_reference": 28,
        "029Animal_Wild animals_Rodents_ rats_ mice_Mouse___3Samples_reference": 29,
        "030Animal_Wild animals_Snake_Rattle_reference": 30,
        "031Channel_ environment and background_Noise_Background noise_Mains hum_reference": 31,
        "032Channel_ environment and background_Noise_Background noise_Tape hiss_reference": 32,
        "033Channel_ environment and background_Noise_Pink noise_reference": 33,
        "034Channel_ environment and background_Noise_White noise_reference": 34,
        "035Human sounds_Digestive_Biting_ Chewing_reference": 35,
        "036Human sounds_Digestive_Burping_ eructation_reference": 36,
        "037Human sounds_Digestive_Fart_reference": 37,
        "038Human sounds_Digestive_Gargling_reference": 38,
        "039Human sounds_Digestive_Hiccup_reference": 39,
        "040Human sounds_Digestive_Stomach rumble_reference": 40,
        "041Human sounds_Hands_Clapping_reference": 41,
        "042Human sounds_Hands_Finger snapping_reference": 42,
        "043Human sounds_Heart sounds_ heartbeat_reference": 43,
        "044Human sounds_Human group actions_Applause_reference": 44,
        "045Human sounds_Human group actions_Booing_reference": 45,
        "046Human sounds_Human group actions_Cheering_reference": 46,
        "047Human sounds_Human group actions_Children shouting_reference": 47,
        "048Human sounds_Human locomotion_Run_reference": 48,
        "049Human sounds_Human locomotion_Shuffle_reference": 49,
        "050Human sounds_Human locomotion_Walk_ footsteps_reference": 50,
        "051Human sounds_Human voice_Crying_ sobbing_Baby cry_ infant cry_reference": 51,
        "052Human sounds_Human voice_Groan_reference": 52,
        "053Human sounds_Human voice_Grunt_reference": 53,
        "054Human sounds_Human voice_Humming_reference": 54,
        "055Human sounds_Human voice_Laughter_Baby laughter_reference": 55,
        "056Human sounds_Human voice_Laughter_Chuckle_ chortle_reference": 56,
        "057Human sounds_Human voice_Laughter_Giggle_reference": 57,
        "058Human sounds_Human voice_Screaming_reference": 58,
        "059Human sounds_Human voice_Shout_Children shouting_reference": 59,
        "060Human sounds_Human voice_Shout_Whoop_reference": 60,
        "061Human sounds_Human voice_Shout_Yell_reference": 61,
        "062Human sounds_Human voice_Sigh_reference": 62,
        "063Human sounds_Human voice_Wail_ moan_reference": 63,
        "064Human sounds_Human voice_Whispering_reference": 64,
        "065Human sounds_Human voice_Yawn_reference": 65,
        "066Human sounds_Respiratory sounds_Breathing_Gasp_reference": 66,
        "067Human sounds_Respiratory sounds_Breathing_Pant_reference": 67,
        "068Human sounds_Respiratory sounds_Breathing_Snoring_reference": 68,
        "069Human sounds_Respiratory sounds_Cough_Throat clearing_reference": 69,
        "070Human sounds_Respiratory sounds_Sneeze_reference": 70,
        "071Human sounds_Respiratory sounds_Sniff_reference": 71,
        "072Human sounds_Whistling_Wolf-whistling_reference": 72,
        "073Music_Musical instrument_Accordion_reference": 73,
        "074Music_Musical instrument_Bagpipes_reference": 74,
        "075Music_Musical instrument_Bell_Bicycle bell_reference": 75,
        "076Music_Musical instrument_Bell_Chime_Wind chime_reference": 76,
        "077Music_Musical instrument_Bell_Church bell_reference": 77,
        "078Music_Musical instrument_Bell_Jingle bell_reference": 78,
        "079Music_Musical instrument_Bell_Tuning fork_reference": 79,
        "080Music_Musical instrument_Bowed string instrument_Cello_reference": 80,
        "081Music_Musical instrument_Bowed string instrument_Double bass_reference": 81,
        "082Music_Musical instrument_Bowed string instrument_String section_reference": 82,
        "083Music_Musical instrument_Bowed string instrument_Violin_ fiddle_Pizzicato_reference": 83,
        "084Music_Musical instrument_Brass instrument_Bugle_reference": 84,
        "085Music_Musical instrument_Brass instrument_Cornet_reference": 85,
        "086Music_Musical instrument_Brass instrument_French horn_reference": 86,
        "087Music_Musical instrument_Brass instrument_Trombone_reference": 87,
        "088Music_Musical instrument_Brass instrument_Trumpet_reference": 88,
        "089Music_Musical instrument_Choir_reference": 89,
        "090Music_Musical instrument_Didgeridoo_reference": 90,
        "091Music_Musical instrument_Harmonica_reference": 91,
        "092Music_Musical instrument_Harp_reference": 92,
        "093Music_Musical instrument_Keyboard (musical)_Harpsichord_reference": 93,
        "094Music_Musical instrument_Keyboard (musical)_Organ_Electronic organ_reference": 94,
        "095Music_Musical instrument_Keyboard (musical)_Organ_Hammond organ_reference": 95,
        "096Music_Musical instrument_Keyboard (musical)_Piano_Electric piano_reference": 96,
        "097Music_Musical instrument_Keyboard (musical)_Synthesizer_Mellotron_reference": 97,
        "098Music_Musical instrument_Keyboard (musical)_Synthesizer_Sampler_reference": 98,
        "099Music_Musical instrument_Musical ensemble_reference": 99,
        "100Music_Musical instrument_Orchestra_reference": 100,
        "101Music_Musical instrument_Percussion_Cowbell_reference": 101,
        "102Music_Musical instrument_Percussion_Cymbal_Crash Cymbal_reference": 102,
        "103Music_Musical instrument_Percussion_Cymbal_Hi-hat_reference": 103,
        "104Music_Musical instrument_Percussion_Drum kit_Drum machine_reference": 104,
        "105Music_Musical instrument_Percussion_Drum_Bass drum_reference": 105,
        "106Music_Musical instrument_Percussion_Drum_Snare drum_Drum roll_reference": 106,
        "107Music_Musical instrument_Percussion_Drum_Snare drum_Rimshot_reference": 107,
        "108Music_Musical instrument_Percussion_Drum_Tabla_reference": 108,
        "109Music_Musical instrument_Percussion_Drum_Timpani_reference": 109,
        "110Music_Musical instrument_Percussion_Gong_reference": 110,
        "111Music_Musical instrument_Percussion_Mallet percussion_Glockenspiel_reference": 111,
        "112Music_Musical instrument_Percussion_Mallet percussion_Marimba_ xylophone_reference": 112,
        "113Music_Musical instrument_Percussion_Mallet percussion_Vibraphone_reference": 113,
        "114Music_Musical instrument_Percussion_Rattle (instrument)_Maraca_reference": 114,
        "115Music_Musical instrument_Percussion_Tambourine_reference": 115,
        "116Music_Musical instrument_Percussion_Tubular bells_reference": 116,
        "117Music_Musical instrument_Percussion_Wood block_reference": 117,
        "118Music_Musical instrument_Plucked string instrument_Banjo_reference": 118,
        "119Music_Musical instrument_Plucked string instrument_Guitar_Acoustic guitar_reference": 119,
        "120Music_Musical instrument_Plucked string instrument_Guitar_Bass guitar_reference": 120,
        "121Music_Musical instrument_Plucked string instrument_Guitar_Electric guitar_reference": 121,
        "122Music_Musical instrument_Plucked string instrument_Guitar_Steel guitar_ slide guitar_reference": 122,
        "123Music_Musical instrument_Plucked string instrument_Guitar_Strum_reference": 123,
        "124Music_Musical instrument_Plucked string instrument_Guitar_Tapping (guitar technique)_reference": 124,
        "125Music_Musical instrument_Plucked string instrument_Mandolin_reference": 125,
        "126Music_Musical instrument_Plucked string instrument_Sitar_reference": 126,
        "127Music_Musical instrument_Plucked string instrument_Ukulele_reference": 127,
        "128Music_Musical instrument_Plucked string instrument_Zither_reference": 128,
        "129Music_Musical instrument_Scratching (performance technique)_reference": 129,
        "130Music_Musical instrument_Singing bowl_reference": 130,
        "131Music_Musical instrument_Theremin_reference": 131,
        "132Music_Musical instrument_Wind instrument_ woodwind instrument_Bassoon_reference": 132,
        "133Music_Musical instrument_Wind instrument_ woodwind instrument_Clarinet_reference": 133,
        "134Music_Musical instrument_Wind instrument_ woodwind instrument_Flute_reference": 134,
        "135Music_Musical instrument_Wind instrument_ woodwind instrument_Oboe_reference": 135,
        "136Music_Musical instrument_Wind instrument_ woodwind instrument_Saxophone_Alto saxophone_reference": 136,
        "137Music_Musical instrument_Wind instrument_ woodwind instrument_Saxophone_Soprano saxophone_reference": 137,
        "138Natural sounds_Fire_Crackle_reference": 138,
        "139Natural sounds_Thunderstorm_Thunder_reference": 139,
        "140Natural sounds_Water_Gurgling_reference": 140,
        "141Natural sounds_Water_Ocean_Waves_ surf_reference": 141,
        "142Natural sounds_Water_Rain_Rain on surface_reference": 142,
        "143Natural sounds_Water_Stream_reference": 143,
        "144Natural sounds_Water_Steam_Hiss_reference": 144,
        "145Natural sounds_Water_Waterfall_reference": 145,
        "146Natural sounds_Wind_Howl (wind)_reference": 146,
        "147Natural sounds_Wind_Rustling leaves_reference": 147,
        "148Sounds of things_Alarm_Air horn_ truck horn_reference": 148,
        "149Sounds of things_Alarm_Alarm clock_reference": 149,
        "150Sounds of things_Alarm_Bicycle bell_reference": 150,
        "151Sounds of things_Alarm_Buzzer_reference": 151,
        "152Sounds of things_Alarm_Car alarm_reference": 152,
        "153Sounds of things_Alarm_Doorbell_Ding-dong_reference": 153,
        "154Sounds of things_Alarm_Fire alarm_reference": 154,
        "155Sounds of things_Alarm_Foghorn_reference": 155,
        "156Sounds of things_Alarm_Siren_Ambulance (siren)_reference": 156,
        "157Sounds of things_Alarm_Siren_Civil defense siren_reference": 157,
        "158Sounds of things_Alarm_Siren_Fire engine_ fire truck (siren)_reference": 158,
        "159Sounds of things_Alarm_Siren_Police car (siren)_reference": 159,
        "160Sounds of things_Alarm_Telephone_Busy signal_reference": 160,
        "161Sounds of things_Alarm_Telephone_Cellphone buzz_ vibrating alert_reference": 161,
        "162Sounds of things_Alarm_Telephone_Dial tone_reference": 162,
        "163Sounds of things_Alarm_Telephone_Ringtone_reference": 163,
        "164Sounds of things_Alarm_Telephone_Telephone bell ringing_reference": 164,
        "165Sounds of things_Alarm_Telephone_Telephone dialing_ DTMF_reference": 165,
        "166Sounds of things_Alarm_Vehicle horn_ car horn_ honking_Toot_reference": 166,
        "167Sounds of things_Alarm_Whistle_Kettle whistle_reference": 167,
        "168Sounds of things_Bell_Chime_Wind chime_reference": 168,
        "169Sounds of things_Domestic sounds_ home sounds_Bathtub (filling or washing)_reference": 169,
        "170Sounds of things_Domestic sounds_ home sounds_Blender_reference": 170,
        "171Sounds of things_Domestic sounds_ home sounds_Chopping (food)_reference": 171,
        "172Sounds of things_Domestic sounds_ home sounds_Coin (dropping)_reference": 172,
        "173Sounds of things_Domestic sounds_ home sounds_Cupboard open or close_reference": 173,
        "174Sounds of things_Domestic sounds_ home sounds_Cutlery_ silverware_reference": 174,
        "175Sounds of things_Domestic sounds_ home sounds_Dishes_ pots_ and pans_reference": 175,
        "176Sounds of things_Domestic sounds_ home sounds_Door_Knock_reference": 176,
        "177Sounds of things_Domestic sounds_ home sounds_Door_Slam_reference": 177,
        "178Sounds of things_Domestic sounds_ home sounds_Door_Sliding door_reference": 178,
        "179Sounds of things_Domestic sounds_ home sounds_Door_Squeak_reference": 179,
        "180Sounds of things_Domestic sounds_ home sounds_Drawer open or close_reference": 180,
        "181Sounds of things_Domestic sounds_ home sounds_Electric shaver_ electric razor_reference": 181,
        "182Sounds of things_Domestic sounds_ home sounds_Frying (food)_reference": 182,
        "183Sounds of things_Domestic sounds_ home sounds_Hair dryer_reference": 183,
        "184Sounds of things_Domestic sounds_ home sounds_Keys jangling_reference": 184,
        "185Sounds of things_Domestic sounds_ home sounds_Microwave oven_reference": 185,
        "186Sounds of things_Domestic sounds_ home sounds_Packing tape_ duct tape_reference": 186,
        "187Sounds of things_Domestic sounds_ home sounds_Scissors_reference": 187,
        "188Sounds of things_Domestic sounds_ home sounds_Shuffling cards_reference": 188,
        "189Sounds of things_Domestic sounds_ home sounds_Sink (filling or washing)_reference": 189,
        "190Sounds of things_Domestic sounds_ home sounds_Toilet flush_reference": 190,
        "191Sounds of things_Domestic sounds_ home sounds_Toothbrush_Electric toothbrush_reference": 191,
        "192Sounds of things_Domestic sounds_ home sounds_Typing_Computer keyboard_reference": 192,
        "193Sounds of things_Domestic sounds_ home sounds_Typing_Typewriter_reference": 193,
        "194Sounds of things_Domestic sounds_ home sounds_Vacuum cleaner_reference": 194,
        "195Sounds of things_Domestic sounds_ home sounds_Velcro_ hook and loop fastener_reference": 195,
        "196Sounds of things_Domestic sounds_ home sounds_Water tap_ faucet_reference": 196,
        "197Sounds of things_Domestic sounds_ home sounds_Writing_reference": 197,
        "198Sounds of things_Domestic sounds_ home sounds_Zipper (clothing)_reference": 198,
        "199Sounds of things_Engine_Accelerating_ revving_ vroom_reference": 199,
        "200Sounds of things_Engine_Engine starting_reference": 200,
        "201Sounds of things_Engine_Idling_reference": 201,
        "202Sounds of things_Engine_Jet engine_reference": 202,
        "203Sounds of things_Engine_Light engine (high frequency)_Chainsaw_reference": 203,
        "204Sounds of things_Engine_Light engine (high frequency)_Dental drill_ dentist_s drill_reference": 204,
        "205Sounds of things_Engine_Light engine (high frequency)_Lawn mower_reference": 205,
        "206Sounds of things_Engine_Medium engine (mid frequency)_reference": 206,
        "207Sounds of things_Explosion_Boom_Sonic boom_reference": 207,
        "208Sounds of things_Explosion_Burst_ pop_reference": 208,
        "209Sounds of things_Explosion_Eruption_reference": 209,
        "210Sounds of things_Explosion_Fireworks_Firecracker_reference": 210,
        "211Sounds of things_Explosion_Gunshot_ gunfire_Artillery fire_reference": 211,
        "212Sounds of things_Explosion_Gunshot_ gunfire_Cap gun_reference": 212,
        "213Sounds of things_Explosion_Gunshot_ gunfire_Fusillade_reference": 213,
        "214Sounds of things_Explosion_Gunshot_ gunfire_Machine gun_reference": 214,
        "215Sounds of things_Glass_Chink_ clink_reference": 215,
        "216Sounds of things_Glass_Shatter_reference": 216,
        "217Sounds of things_Liquid_Boiling_reference": 217,
        "218Sounds of things_Liquid_Drip_reference": 218,
        "219Sounds of things_Liquid_Fill (with liquid)_reference": 219,
        "220Sounds of things_Liquid_Pour_Gush_reference": 220,
        "221Sounds of things_Liquid_Pour_Trickle_ dribble_reference": 221,
        "222Sounds of things_Liquid_Pump (liquid)_reference": 222,
        "223Sounds of things_Liquid_Splash_ splatter_Slosh_reference": 223,
        "224Sounds of things_Liquid_Spray_reference": 224,
        "225Sounds of things_Liquid_Squish_reference": 225,
        "226Sounds of things_Liquid_Stir_reference": 226,
        "227Sounds of things_Mechanisms_Air conditioning_reference": 227,
        "228Sounds of things_Mechanisms_Camera_Single-lens reflex camera_reference": 228,
        "229Sounds of things_Mechanisms_Cash register_reference": 229,
        "230Sounds of things_Mechanisms_Clock_Tick_reference": 230,
        "231Sounds of things_Mechanisms_Clock_Tick-tock_reference": 231,
        "232Sounds of things_Mechanisms_Gears_reference": 232,
        "233Sounds of things_Mechanisms_Mechanical fan_reference": 233,
        "234Sounds of things_Mechanisms_Printer_reference": 234,
        "235Sounds of things_Mechanisms_Pulleys_reference": 235,
        "236Sounds of things_Mechanisms_Ratchet_ pawl_reference": 236,
        "237Sounds of things_Mechanisms_Sewing machine_reference": 237,
        "238Sounds of things_Miscellaneous sources_Arrow_Thump_ thud_Clunk_reference": 238,
        "239Sounds of things_Miscellaneous sources_Duck call (hunting tool)_reference": 239,
        "240Sounds of things_Miscellaneous sources_Sonar_reference": 240,
        "241Sounds of things_Specific impact sounds_Basketball bounce_reference": 241,
        "242Sounds of things_Tools_Filing (rasp)_reference": 242,
        "243Sounds of things_Tools_Hammer_reference": 243,
        "244Sounds of things_Tools_Jackhammer_reference": 244,
        "245Sounds of things_Tools_Power tool_Drill_Dental drill_ dentist_s drill_reference": 245,
        "246Sounds of things_Tools_Sanding_reference": 246,
        "247Sounds of things_Tools_Sawing_reference": 247,
        "248Sounds of things_Vehicle_Aircraft_Aircraft engine_Propeller_ airscrew_reference": 248,
        "249Sounds of things_Vehicle_Aircraft_Fixed-wing aircraft_ airplane_reference": 249,
        "250Sounds of things_Vehicle_Aircraft_Helicopter_reference": 250,
        "251Sounds of things_Vehicle_Boat_ Water vehicle_Motorboat_ speedboat_reference": 251,
        "252Sounds of things_Vehicle_Boat_ Water vehicle_Rowboat_ canoe_ kayak_reference": 252,
        "253Sounds of things_Vehicle_Boat_ Water vehicle_Sailboat_ sailing ship_reference": 253,
        "254Sounds of things_Vehicle_Boat_ Water vehicle_Ship_reference": 254,
        "255Sounds of things_Vehicle_Motor vehicle (road)_Bus_reference": 255,
        "256Sounds of things_Vehicle_Motor vehicle (road)_Car_Car alarm_reference": 256,
        "257Sounds of things_Vehicle_Motor vehicle (road)_Car_Car passing by_reference": 257,
        "258Sounds of things_Vehicle_Motor vehicle (road)_Car_Power windows_ electric windows_reference": 258,
        "259Sounds of things_Vehicle_Motor vehicle (road)_Car_Race car_ auto racing_reference": 259,
        "260Sounds of things_Vehicle_Motor vehicle (road)_Car_Skidding_reference": 260,
        "261Sounds of things_Vehicle_Motor vehicle (road)_Car_Vehicle horn_ car horn_ honking_Toot_reference": 261,
        "262Sounds of things_Vehicle_Motor vehicle (road)_Emergency vehicle_Ambulance (siren)_reference": 262,
        "263Sounds of things_Vehicle_Motor vehicle (road)_Emergency vehicle_Fire engine_ fire truck (siren)_reference": 263,
        "264Sounds of things_Vehicle_Motor vehicle (road)_Emergency vehicle_Police car (siren)_reference": 264,
        "265Sounds of things_Vehicle_Motor vehicle (road)_Motorcycle_reference": 265,
        "266Sounds of things_Vehicle_Motor vehicle (road)_Traffic noise_ roadway noise_reference": 266,
        "267Sounds of things_Vehicle_Motor vehicle (road)_Truck_Air brake_reference": 267,
        "268Sounds of things_Vehicle_Motor vehicle (road)_Truck_Air horn_ truck horn_reference": 268,
        "269Sounds of things_Vehicle_Motor vehicle (road)_Truck_Ice cream truck_ ice cream van_reference": 269,
        "270Sounds of things_Vehicle_Motor vehicle (road)_Truck_Reversing beeps_reference": 270,
        "271Sounds of things_Vehicle_Non-motorized land vehicle_Bicycle_Bicycle bell_reference": 271,
        "272Sounds of things_Vehicle_Non-motorized land vehicle_Skateboard_reference": 272,
        "273Sounds of things_Vehicle_Rail transport_Railroad car_ train wagon_reference": 273,
        "274Sounds of things_Vehicle_Rail transport_Subway_ metro_ underground_reference": 274,
        "275Sounds of things_Vehicle_Rail transport_Train wheels squealing_reference": 275,
        "276Sounds of things_Vehicle_Rail transport_Train_Train horn_reference": 276,
        "277Sounds of things_Vehicle_Rail transport_Train_Train whistle_reference": 277,
        "278Sounds of things_Wood_Chop_reference": 278,
        "279Sounds of things_Wood_Crack_reference": 279,
        "280Sounds of things_Wood_Snap_reference": 280,
        "281Sounds of things_Wood_Splinter_reference": 281,
        "282Source-ambiguous sounds_Deformable shell_Crumpling_ crinkling_reference": 282,
        "283Source-ambiguous sounds_Deformable shell_Crushing_reference": 283,
        "284Source-ambiguous sounds_Deformable shell_Tearing_reference": 284,
        "285Source-ambiguous sounds_Generic impact sounds_Bang_reference": 285,
        "286Source-ambiguous sounds_Generic impact sounds_Bouncing_reference": 286,
        "287Source-ambiguous sounds_Generic impact sounds_Breaking_reference": 287,
        "288Source-ambiguous sounds_Generic impact sounds_Flap_reference": 288,
        "289Source-ambiguous sounds_Generic impact sounds_Knock_reference": 289,
        "290Source-ambiguous sounds_Generic impact sounds_Slap_ smack_reference": 290,
        "291Source-ambiguous sounds_Generic impact sounds_Smash_ crash_reference": 291,
        "292Source-ambiguous sounds_Generic impact sounds_Tap_reference": 292,
        "293Source-ambiguous sounds_Generic impact sounds_Thump_ thud_Clunk_reference": 293,
        "294Source-ambiguous sounds_Generic impact sounds_Thump_ thud_Thunk_reference": 294,
        "295Source-ambiguous sounds_Generic impact sounds_Whack_ thwack_reference": 295,
        "296Source-ambiguous sounds_Generic impact sounds_Whip_reference": 296,
        "297Source-ambiguous sounds_Surface contact_Grind_reference": 297,
        "298Source-ambiguous sounds_Surface contact_Roll_reference": 298,
        "299Source-ambiguous sounds_Surface contact_Rub_reference": 299,
        "300Source-ambiguous sounds_Surface contact_Scrape_reference": 300,
        "301Source-ambiguous sounds_Surface contact_Scratch_reference": 301,
    }
    config = TaskConfig(
        encoder=encoder,
        epochs=20,
        k_fold_splits=list(range(0, 3)),
        label_processor=lambda x: class_label_maps[x[task]],
        name="vocalimitations",
        output_dim=len(class_label_maps),
        zenodo_id="14862060",
    )
    config.audio_tar_name_of_split = {
        fold: f"vocal_imitations_fold{fold:02}_0000000.tar" for fold in config.k_fold_splits
    }
    config.encoded_tar_name_of_split = {
        fold: f"vocalimiations-wds-encoded-fold-{fold}-*.tar" for fold in config.k_fold_splits
    }
    return config



from loguru import logger

from xares.task import TaskConfig


def voxlingua33_config(encoder) -> TaskConfig:
    class_label_maps = {
        "uk": 0,
        "zh": 1,
        "nn": 2,
        "hr": 3,
        "nl": 4,
        "fa": 5,
        "el": 6,
        "sv": 7,
        "hy": 8,
        "pl": 9,
        "hu": 10,
        "mk": 11,
        "is": 12,
        "pt": 13,
        "sl": 14,
        "ar": 15,
        "en": 16,
        "lv": 17,
        "ru": 18,
        "az": 19,
        "no": 20,
        "lt": 21,
        "es": 22,
        "tr": 23,
        "fr": 24,
        "de": 25,
        "et": 26,
        "ur": 27,
        "da": 28,
        "fi": 29,
        "it": 30,
        "ja": 31,
        "sr": 32,
    }

    config = TaskConfig(
        batch_size_train=64,
        crop_length=10,
        encoder=encoder,
        epochs=10,
        label_processor=lambda x: class_label_maps[x["labels"]],
        learning_rate=1e-3,
        name="voxlingua33",
        output_dim=len(class_label_maps),
        test_split="dev",
        train_split="train_subset",
        valid_split="dev",
        zenodo_id="14812245",
    )

    if config.use_mini_dataset:
        logger.warning(f"Dataset {config.name} uses mini version for faster evaluation.")
        config.audio_tar_name_of_split[config.train_split] = "train_subset_2k_00000{00..01}.tar"

    return config



from xares.task import TaskConfig


def fma_genre_config(encoder) -> TaskConfig:
    data_key = "genre"
    class_label_maps = {
        "Hip-Hop": 0,
        "Pop": 1,
        "Folk": 2,
        "Experimental": 3,
        "Rock": 4,
        "International": 5,
        "Electronic": 6,
        "Instrumental": 7,
    }
    return TaskConfig(
        crop_length=10,  # 10s
        encoder=encoder,
        epochs=3,
        label_processor=lambda x: class_label_maps[x[data_key]],
        name="freemusicarchive",
        output_dim=len(class_label_maps),
        test_split="fma_small_test",
        train_split="fma_small_train",
        valid_split="fma_small_valid",
        zenodo_id="14725056",
    )



from xares.task import TaskConfig


def vocalsound_config(encoder) -> TaskConfig:
    return TaskConfig(
        encoder=encoder,
        label_processor=lambda x: {
            "laughter": 0,
            "sigh": 1,
            "throatclearing": 2,
            "cough": 3,
            "sneeze": 4,
            "sniff": 5,
        }.get(x["label"], -1),
        name="vocalsound",
        output_dim=6,
        zenodo_id="14722710",
    )



# Reduced from `xiaomitts/common/audiowebdataset.py`

import json
import warnings
from functools import partial
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Literal, Optional, Sequence, Tuple, Union  # type: ignore

import numpy as np
import torch
import torchaudio
import webdataset as wds
from loguru import logger


def fast_warn_and_continue(exn):
    if "Format not recognised" in repr(exn):
        raise RuntimeError("Format not recognised")

    warnings.warn(repr(exn))
    return True


def crop_or_pad_audio(wav: torch.Tensor, crop_size: int, pad_last: bool = False):
    n_samples, *_ = wav.shape
    available_crops = n_samples // crop_size
    for i in range(available_crops):
        crop = wav[i * crop_size : (i + 1) * crop_size, ...]
        yield crop

    if (available_crops == 0) and pad_last:
        last_crop = wav[available_crops * crop_size :, ...]
        padded = torch.zeros((crop_size, *last_crop.shape[1:]))
        padded[: last_crop.shape[0]] = last_crop
        yield padded


def _seq_crop_audio(
    data,
    crop_length: None | float,
    mono: bool = True,
    drop_clipped: bool = True,
    pad_last: bool = False,
    handler=None,
):
    """WebDataset crop filter, yields sequential crops"""
    for sample in data:
        audio, *extra = sample
        audio, sr = audio
        if mono and audio.ndim == 2:
            audio = audio.mean(0)
        if audio.abs().max() >= 0.99 and drop_clipped:
            continue
        if crop_length is not None:
            crops = crop_or_pad_audio(audio.float(), crop_size=(crop_length * sr), pad_last=pad_last)
        else:
            crops = [audio.float()]

        for crop in crops:
            yield (crop, *extra)


class Audiowebdataset(wds.DataPipeline):

    def __init__(
        self,
        urls,
        tar_shuffle: None | int = None,
        resample: bool = False,
        target_sample_rate: None | int = None,
        batch_size: None | int = None,
        filter_function: None | Callable = None,
        rename_keys: Dict[str, str] = dict(audio="flac;mp3;sox;wav;m4a;ogg;wma", filename="__key__"),
        map_kwargs: None | Dict[str, Callable] = None,
        merge_function: (
            None | Callable
        ) = None,  # merge function is called before batching. In the merge function we can operate on the data in form of a tuple
        handler=fast_warn_and_continue,
    ):
        pipeline: List = [wds.ResampledShards(urls) if resample else wds.SimpleShardList(urls)]

        if tar_shuffle is not None:
            # Tar wise shuffle
            pipeline.extend(
                [
                    wds.detshuffle(
                        bufsize=tar_shuffle,
                        initial=tar_shuffle // 4,
                    ),
                    wds.split_by_node,
                    wds.split_by_worker,
                    # at this point, we have an iterator over the shards assigned to each worker at each node
                    wds.tarfile_to_samples(handler=handler),
                    wds.shuffle(
                        bufsize=tar_shuffle,
                        initial=tar_shuffle // 4,
                    ),
                ]
            )
        else:
            pipeline.extend([wds.split_by_node, wds.split_by_worker, wds.tarfile_to_samples(handler=handler)])

        # Decode i.e., bytes object to a python-accessible obj.
        pipeline.extend([wds.decode(wds.torch_audio, handler=handler), wds.rename(**rename_keys, handler=handler)])

        if map_kwargs:
            pipeline.extend([wds.map_dict(**map_kwargs)])
        # Filter function takes a sample (key: value) as input and returns True for valid samples, otherwise false
        if filter_function:
            pipeline.extend([wds.select(filter_function)])

        # Resample audio, useful when dataset is not monotonous in sampling rate
        if target_sample_rate:
            assert "audio" in rename_keys.keys(), "target_sample_rate requires key_maps=dict(audio='flac;mp3;wav')"

            def resample_audio(audio_sr: Tuple[torch.Tensor, int]) -> Tuple[torch.Tensor, int]:
                audio, sr = audio_sr
                audio = torchaudio.functional.resample(audio, sr, target_sample_rate)
                return (audio, target_sample_rate)

            pipeline.extend([wds.map_dict(audio=resample_audio)])

        # Webdataset support batching and parallel reading using
        # num_workers only with tuples, not dicts
        pipeline.extend(
            [
                wds.to_tuple(*rename_keys.keys()),
            ]
        )

        if merge_function is not None:
            pipeline.extend([merge_function])

        if batch_size is not None:
            pipeline.append(
                wds.batched(
                    batch_size,
                    collation_fn=partial(
                        wds.filters.default_collation_fn, combine_tensors=False, combine_scalars=False
                    ),
                )
            )

        super().__init__(pipeline)


# Can also replace with wds.Randomix
class BalancedDatasetSampler(wds.DataPipeline, wds.compat.FluidInterface):

    def __init__(self, **datasets):

        super().__init__()
        self.datasets = datasets

    def __iter__(self):
        sources = {k: iter(ds) for k, ds in self.datasets.items()}
        while True:
            for k, source in sources.items():
                try:
                    yield next(source)
                except StopIteration:
                    break


def expand_with_brace(lists: Iterable[str] | str):
    import braceexpand

    r = []
    for l in lists:
        if "*" in l:
            # Expand using "posix" based *
            r.extend([str(f) for f in Path(l).parent.glob(Path(l).name)])
        else:
            r.extend(braceexpand.braceexpand(l))
        if any(not Path(l).exists() for l in r):
            raise FileNotFoundError(f"One of the paths does not exist: {r}")
    return r


def pad(tensorlist: Sequence[torch.Tensor], padding_value: float = 0.0):
    # Tensors are expected to be B, ..., T
    lengths = [f.shape[-1] for f in tensorlist]
    dims = tensorlist[0].shape
    trailing_dims = dims[:-1]
    batch_dim = len(lengths)
    num_raw_samples = max(lengths)
    out_dims = (batch_dim,) + trailing_dims + (num_raw_samples,)
    out_tensor = torch.full(out_dims, fill_value=padding_value, dtype=tensorlist[0].dtype)
    for i, tensor in enumerate(tensorlist):
        length = tensor.shape[-1]
        out_tensor[i, ..., :length] = tensor[..., :length]
    return out_tensor, torch.as_tensor(lengths)


def collate_with_lengths_wds(
    samples: List[Iterable], combine_scalars: bool = True, flatten: bool = True, combine_tensors: bool = True
):
    batched = list(zip(*samples))
    result = []
    for i, b in enumerate(batched):
        if isinstance(b[0], (int, float)):
            if combine_scalars:
                b = np.array(list(b))
        elif isinstance(b[0], torch.Tensor):
            if combine_tensors:
                b = pad(list(b))
        elif isinstance(b[0], np.ndarray):
            if combine_tensors:
                b = np.array(list(b))
        else:
            b = list(b)
        # Do not flatten lists, i.e., some filenames
        if flatten and not isinstance(b, list):
            result.extend(b)
        else:
            result.append(b)
    return result


# Returns (single) dicts with (audio=audio_data, *extra ), useful for only reading audio and keeping other items the same
def create_rawaudio_webdataset(
    urls: List[str] | Dict[str, List[str]],
    target_sample_rate: Optional[int] = None,
    mono: bool = True,
    num_workers: int = 4,
    batch_size: int = 64,
    crop_length: float | None = None,
    pad_last: bool = False,  # If only 1 crop available, use padding
    **kwargs,
):

    dataset_kwargs = dict(
        batch_size=batch_size,
        rename_keys=(dict(audio="flac;mp3;sox;wav;m4a;ogg;wma", json="json", filename="__key__")),
        target_sample_rate=target_sample_rate,
        merge_function=partial(
            _seq_crop_audio, crop_length=crop_length, mono=mono, drop_clipped=False, pad_last=pad_last
        ),
    )
    urls = expand_with_brace(urls)
    dataset = Audiowebdataset(urls, **dataset_kwargs)
    # Set num_workers at most to number of tars, otherwise some processes will do nothing, slowing down dataloading
    dataloader = wds.WebLoader(dataset, num_workers=min(len(urls), num_workers), batch_size=None).unbatched()
    dataloader = dataloader.batched(
        batch_size,
        collation_fn=partial(collate_with_lengths_wds, flatten=False),
    )
    return dataloader


def create_embedding_webdataset(
    urls: Union[List[str], Dict[str, List[str]]],
    tar_shuffle: None | int = None,
    batch_size: int = 16,
    balanced_sampler: None | bool = False,
    num_workers: int = 4,
    training: bool = False,
    label_processor: None | Callable = None,
    merge_processor: None | Callable = None,
    **kwargs,
):

    dataset_kwargs = dict(
        tar_shuffle=tar_shuffle,
        batch_size=batch_size,
        rename_keys=dict(embedding="npy", target="json", filename="__key__"),
        map_kwargs=dict(
            embedding=lambda x: x.transpose(),
            target=label_processor if label_processor else lambda x: x,
        ),  # Transpose (B,T,D) -> (B,D,T), map the labels if provided
        merge_function=merge_processor,
    )
    if balanced_sampler:
        assert isinstance(urls, dict)
        ds = {k: Audiowebdataset(expand_with_brace(train_data), **dataset_kwargs) for k, train_data in urls.items()}
        dataset = BalancedDatasetSampler(**ds)
    else:
        assert isinstance(urls, list)
        dataset = Audiowebdataset(expand_with_brace(urls), **dataset_kwargs)

    dataloader = wds.WebLoader(
        dataset,
        batch_size=None,
        num_workers=num_workers,
    ).unbatched()
    if training:
        dataloader = dataloader.shuffle(512)
    dataloader = dataloader.batched(
        batch_size,
        collation_fn=partial(collate_with_lengths_wds, flatten=False),
    )
    return dataloader


def write_audio_tar(
    audio_paths: List[str],
    labels: List,
    tar_path: str,
    suffix: str = "wav",
    num_shards: int = 20,
    force: bool = False,
    min_length: int = 100,
):
    assert len(audio_paths) == len(labels), "Number of audio files and labels must match."

    assert len(audio_paths) >= num_shards, "Number of shards must be less than number of audio files."
    shard_size = (len(audio_paths) + num_shards - 1) // num_shards

    def make_sample(filename, label=None):
        with open(filename, "rb") as buf:
            raw_data = buf.read()
        fpath = Path(filename)
        stem_name = str(fpath.stem).replace(".", "_")
        suffix = fpath.suffix.replace(".", "")
        ret_data = {
            suffix: raw_data,
            "__key__": f"{stem_name}",  # Just cast to str
        }
        # If we have some labels, also dump a .json file
        if label is not None:
            if isinstance(label, dict):
                ret_data["json"] = json.dumps(label).encode("utf-8")
            elif isinstance(label, str):
                ret_data["json"] = json.dumps({"label": label}).encode("utf-8")
            else:
                raise ValueError("Label must be either dict or str.")
        return ret_data

    for shard in range(num_shards):
        start_index = shard * shard_size
        end_index = start_index + shard_size

        shard_audio_paths = audio_paths[start_index:end_index]
        shard_labels = labels[start_index:end_index]

        sharded_tar_path = tar_path.replace("*", f"0{shard:05d}")
        if not force and Path(sharded_tar_path).exists():
            logger.info(f"Tar file {sharded_tar_path} already exists.")
            continue

        with wds.TarWriter(sharded_tar_path) as ostream:
            for audio_path, label in zip(shard_audio_paths, shard_labels):
                sample = make_sample(audio_path, label)
                if len(sample[suffix]) < min_length:
                    logger.warning(f"Skipping {audio_path} due to short length.")
                    continue
                ostream.write(sample)



from xares.task import TaskConfig


def speechcommandsv1_config(encoder) -> TaskConfig:
    class_label_maps = {
        "sheila": 0,
        "up": 1,
        "bird": 2,
        "right": 3,
        "left": 4,
        "six": 5,
        "yes": 6,
        "on": 7,
        "one": 8,
        "off": 9,
        "zero": 10,
        "marvin": 11,
        "seven": 12,
        "wow": 13,
        "five": 14,
        "down": 15,
        "three": 16,
        "nine": 17,
        "no": 18,
        "happy": 19,
        "cat": 20,
        "go": 21,
        "bed": 22,
        "house": 23,
        "stop": 24,
        "four": 25,
        "tree": 26,
        "dog": 27,
        "two": 28,
        "eight": 29,
    }
    data_key = "labels"

    return TaskConfig(
        batch_size_train=64,
        encoder=encoder,
        label_processor=lambda x: class_label_maps[x[data_key]],
        learning_rate=1e-3,
        name="speechcommandsv1",
        output_dim=len(class_label_maps),
        test_split="wds-audio-test",
        train_split="wds-audio-train",
        valid_split="wds-audio-valid",
        zenodo_id="14722647",
    )



import torch
from loguru import logger


def check_audio_encoder(encoder: torch.nn.Module):
    if not type(encoder).__name__.endswith("Encoder"):
        logger.error("Class name must end with 'Encoder'")
        return False

    if not isinstance(encoder, torch.nn.Module):
        logger.error(f"Expected torch.nn.Module for encoder, got {type(encoder)}")
        return False

    if not hasattr(encoder, "output_dim"):
        logger.error("Encoder must have a 'output_dim' attribute")
        return False

    if not hasattr(encoder, "sampling_rate"):
        logger.error("Encoder must have a 'sampling_rate' attribute")
        return False

    if not hasattr(encoder, "hop_size_in_ms"):
        logger.error("Encoder must have a 'hop_size_in_ms' attribute")
        print(
            """
            hop_size_in_ms determines the time shift (in milliseconds) between consecutive frames in the encoded output.

            For example:
            - If input size is (B=1, T=50000) and output size is (B=1, T'=78, D), with a sampling rate of 16000 Hz:
            - Input duration = 50000 / 16000 = 3125 ms.
            - hop_size_in_ms ≈ 3125 / 78 ≈ 40 ms.
            - This means the encoder shifts 40 ms forward for each frame.

            It directly impacts the relationship between input length and output frames.
            """
        )
        return False

    sample_audio = torch.randn(2, 50000)
    try:
        encoded_audio = encoder(sample_audio)
    except Exception as e:
        logger.error(f"Failed to encode the sample audio: {e}")
        return False

    if not isinstance(encoded_audio, torch.Tensor):
        logger.error(f"Expected tensor for encoded_audio, got {type(encoded_audio)}")
        return False

    if encoded_audio.ndim != 3:  # [B, T, D]
        logger.error(f"Expected 3D tensor [B, T, D] for encoded_audio, got {encoded_audio.dim()}D tensor")
        return False

    if encoded_audio.size(0) != sample_audio.size(0):
        logger.error(f"Expected batch size={sample_audio.size(0)} for encoded_audio, got {encoded_audio.size(0)}")
        return False

    if encoded_audio.size(2) != encoder.output_dim:
        logger.error(f"Expected output_dim={encoder.output_dim} for encoded_audio, got {encoded_audio.size(2)}")
        return False

    return True



from loguru import logger

from xares.task import TaskConfig


def librispeech_male_female_config(encoder) -> TaskConfig:
    config = TaskConfig(
        crop_length=5,
        encoder=encoder,
        label_processor=lambda x: 0 if x["gender"] == "M" else 1,
        name="librispeechmalefemale",
        output_dim=2,
        test_split="test-clean",
        train_split="train-clean-100",
        valid_split="dev-clean",
        zenodo_id="14716252",
    )

    if config.use_mini_dataset:
        logger.warning(f"Dataset {config.name} uses mini version for faster evaluation.")
        config.audio_tar_name_of_split[config.train_split] = "train-clean-100-000000.tar"

    return config



from typing import Any, Dict

import torch

from xares.task import TaskConfig


def fsd50k_config(encoder) -> TaskConfig:
    data_key = "soundevent"
    class_label_maps = {
        "Accelerating_and_revving_and_vroom": 0,
        "Accordion": 1,
        "Acoustic_guitar": 2,
        "Aircraft": 3,
        "Alarm": 4,
        "Animal": 5,
        "Applause": 6,
        "Bark": 7,
        "Bass_drum": 8,
        "Bass_guitar": 9,
        "Bathtub_(filling_or_washing)": 10,
        "Bell": 11,
        "Bicycle": 12,
        "Bicycle_bell": 13,
        "Bird": 14,
        "Bird_vocalization_and_bird_call_and_bird_song": 15,
        "Boat_and_Water_vehicle": 16,
        "Boiling": 17,
        "Boom": 18,
        "Bowed_string_instrument": 19,
        "Brass_instrument": 20,
        "Breathing": 21,
        "Burping_and_eructation": 22,
        "Bus": 23,
        "Buzz": 24,
        "Camera": 25,
        "Car": 26,
        "Car_passing_by": 27,
        "Cat": 28,
        "Chatter": 29,
        "Cheering": 30,
        "Chewing_and_mastication": 31,
        "Chicken_and_rooster": 32,
        "Child_speech_and_kid_speaking": 33,
        "Chime": 34,
        "Chink_and_clink": 35,
        "Chirp_and_tweet": 36,
        "Chuckle_and_chortle": 37,
        "Church_bell": 38,
        "Clapping": 39,
        "Clock": 40,
        "Coin_(dropping)": 41,
        "Computer_keyboard": 42,
        "Conversation": 43,
        "Cough": 44,
        "Cowbell": 45,
        "Crack": 46,
        "Crackle": 47,
        "Crash_cymbal": 48,
        "Cricket": 49,
        "Crow": 50,
        "Crowd": 51,
        "Crumpling_and_crinkling": 52,
        "Crushing": 53,
        "Crying_and_sobbing": 54,
        "Cupboard_open_or_close": 55,
        "Cutlery_and_silverware": 56,
        "Cymbal": 57,
        "Dishes_and_pots_and_pans": 58,
        "Dog": 59,
        "Domestic_animals_and_pets": 60,
        "Domestic_sounds_and_home_sounds": 61,
        "Door": 62,
        "Doorbell": 63,
        "Drawer_open_or_close": 64,
        "Drill": 65,
        "Drip": 66,
        "Drum": 67,
        "Drum_kit": 68,
        "Electric_guitar": 69,
        "Engine": 70,
        "Engine_starting": 71,
        "Explosion": 72,
        "Fart": 73,
        "Female_singing": 74,
        "Female_speech_and_woman_speaking": 75,
        "Fill_(with_liquid)": 76,
        "Finger_snapping": 77,
        "Fire": 78,
        "Fireworks": 79,
        "Fixed-wing_aircraft_and_airplane": 80,
        "Fowl": 81,
        "Frog": 82,
        "Frying_(food)": 83,
        "Gasp": 84,
        "Giggle": 85,
        "Glass": 86,
        "Glockenspiel": 87,
        "Gong": 88,
        "Growling": 89,
        "Guitar": 90,
        "Gull_and_seagull": 91,
        "Gunshot_and_gunfire": 92,
        "Gurgling": 93,
        "Hammer": 94,
        "Hands": 95,
        "Harmonica": 96,
        "Harp": 97,
        "Hi-hat": 98,
        "Hiss": 99,
        "Human_group_actions": 100,
        "Human_voice": 101,
        "Idling": 102,
        "Insect": 103,
        "Keyboard_(musical)": 104,
        "Keys_jangling": 105,
        "Knock": 106,
        "Laughter": 107,
        "Liquid": 108,
        "Livestock_and_farm_animals_and_working_animals": 109,
        "Male_singing": 110,
        "Male_speech_and_man_speaking": 111,
        "Mallet_percussion": 112,
        "Marimba_and_xylophone": 113,
        "Mechanical_fan": 114,
        "Mechanisms": 115,
        "Meow": 116,
        "Microwave_oven": 117,
        "Motor_vehicle_(road)": 118,
        "Motorcycle": 119,
        "Music": 120,
        "Musical_instrument": 121,
        "Ocean": 122,
        "Organ": 123,
        "Packing_tape_and_duct_tape": 124,
        "Percussion": 125,
        "Piano": 126,
        "Plucked_string_instrument": 127,
        "Pour": 128,
        "Power_tool": 129,
        "Printer": 130,
        "Purr": 131,
        "Race_car_and_auto_racing": 132,
        "Rail_transport": 133,
        "Rain": 134,
        "Raindrop": 135,
        "Ratchet_and_pawl": 136,
        "Rattle": 137,
        "Rattle_(instrument)": 138,
        "Respiratory_sounds": 139,
        "Ringtone": 140,
        "Run": 141,
        "Sawing": 142,
        "Scissors": 143,
        "Scratching_(performance_technique)": 144,
        "Screaming": 145,
        "Screech": 146,
        "Shatter": 147,
        "Shout": 148,
        "Sigh": 149,
        "Singing": 150,
        "Sink_(filling_or_washing)": 151,
        "Siren": 152,
        "Skateboard": 153,
        "Slam": 154,
        "Sliding_door": 155,
        "Snare_drum": 156,
        "Sneeze": 157,
        "Speech": 158,
        "Speech_synthesizer": 159,
        "Splash_and_splatter": 160,
        "Squeak": 161,
        "Stream": 162,
        "Strum": 163,
        "Subway_and_metro_and_underground": 164,
        "Tabla": 165,
        "Tambourine": 166,
        "Tap": 167,
        "Tearing": 168,
        "Telephone": 169,
        "Thump_and_thud": 170,
        "Thunder": 171,
        "Thunderstorm": 172,
        "Tick": 173,
        "Tick-tock": 174,
        "Toilet_flush": 175,
        "Tools": 176,
        "Traffic_noise_and_roadway_noise": 177,
        "Train": 178,
        "Trickle_and_dribble": 179,
        "Truck": 180,
        "Trumpet": 181,
        "Typewriter": 182,
        "Typing": 183,
        "Vehicle": 184,
        "Vehicle_horn_and_car_horn_and_honking": 185,
        "Walk_and_footsteps": 186,
        "Water": 187,
        "Water_tap_and_faucet": 188,
        "Waves_and_surf": 189,
        "Whispering": 190,
        "Whoosh_and_swoosh_and_swish": 191,
        "Wild_animals": 192,
        "Wind": 193,
        "Wind_chime": 194,
        "Wind_instrument_and_woodwind_instrument": 195,
        "Wood": 196,
        "Writing": 197,
        "Yell": 198,
        "Zipper_(clothing)": 199,
    }

    def manyhot_label_map(x: Dict[str, Any]):
        target_ids = [class_label_maps[item] for item in x[data_key].split(";")]
        target_vec = torch.zeros(len(class_label_maps))
        target_vec.scatter_(0, torch.tensor(target_ids), 1)
        return target_vec.tolist()

    task_config = TaskConfig(
        batch_size_train=64,
        criterion="BCEWithLogitsLoss",
        crop_length=10,  # 10s, even though samples are 30s
        do_knn=False,
        encoder=encoder,
        label_processor=manyhot_label_map,
        metric="mAP",
        name="fsd50k",
        output_dim=len(class_label_maps),
        test_split="fsd50k_eval",
        train_split="fsd50k_train",
        valid_split="fsd50k_validation",
        zenodo_id="14868441",
    )
    return task_config



from math import ceil

import torch

from xares.task import TaskConfig


def desed_config(encoder) -> TaskConfig:
    # We train DESED somewhat different from the standard approach:
    # We use smear out all weak labels across the time axis
    # Then just use a frame-level estimator
    class_label_maps = {
        "Cat": 0,
        "Speech": 1,
        "Running_water": 2,
        "Dishes": 3,
        "Electric_shaver_toothbrush": 4,
        "Vacuum_cleaner": 5,
        "Blender": 6,
        "Alarm_bell_ringing": 7,
        "Dog": 8,
        "Frying": 9,
    }

    def sec_to_frames(sec):
        return ceil(sec / (encoder.hop_size_in_ms / 1000))

    def label_transform(samples):
        target = torch.zeros(len(class_label_maps), sec_to_frames(10), dtype=torch.float32)
        for sample in samples:
            start, end, labelnames = sample["onset"], sample["offset"], sample["label"]
            start = sec_to_frames(start)
            end = sec_to_frames(end)
            for label_name in labelnames.split(","):
                target[class_label_maps[label_name], start:end] = 1.0
        return target

    return TaskConfig(
        criterion="BCEWithLogitsLoss",
        do_knn=False,
        encoder=encoder,
        label_processor=label_transform,
        metric_args=dict(hop_size_in_ms=encoder.hop_size_in_ms if encoder else 0, segment_length_in_s=1.0),
        metric="segmentf1",
        name="desed",
        output_dim=len(class_label_maps),
        task_type="frame",
        test_split="wds-audio-eval",
        train_split="wds-audio-train",
        valid_split="wds-audio-eval",
        zenodo_id="14808180",
    )



import os
from typing import List, Optional, Tuple

import torch
import torch.nn as nn
from jiwer import wer
from torch import Tensor
from transformers import AutoModelForCausalLM, AutoTokenizer

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class EncoderAdapter(nn.Module):
    def __init__(self, dim, intermediate_size, n_embd) -> None:
        super().__init__()
        self.fc_1 = nn.Linear(dim, intermediate_size, bias=True)
        self.fc_2 = nn.Linear(dim, intermediate_size, bias=True)
        self.proj = nn.Linear(intermediate_size, n_embd, bias=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_fc_1 = self.fc_1(x)
        x_fc_2 = self.fc_2(x)
        x = torch.nn.functional.silu(x_fc_1) * x_fc_2
        return self.proj(x)


class Decoder(nn.Module):
    def __init__(self, audio_features_dim: int):
        super().__init__()

        model_name = "Qwen/Qwen2.5-0.5B"
        self.transformer = AutoModelForCausalLM.from_pretrained(model_name)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)

        self.encoder_adapter = EncoderAdapter(
            dim=audio_features_dim,
            intermediate_size=self.transformer.config.intermediate_size,
            n_embd=self.transformer.config.hidden_size,
        )

        self.max_seq_length = 4096
        self.past_key_values = None

    def freeze_lm(self) -> None:
        for param in self.transformer.parameters():
            param.requires_grad = False

    def clear_kvcache(self) -> None:
        self.past_key_values = None

    def forward(
        self,
        encoded_audio: Optional[torch.Tensor] = None,
        attention_mask_a: Optional[torch.Tensor] = None,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask_t: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        enable_kvcache: bool = False,
    ) -> Tuple[List[Tensor], Tensor]:
        assert encoded_audio is not None or input_ids is not None, "Either encoded_audio or input_ids must be provided"
        assert self.transformer.device == self.encoder_adapter.fc_1.weight.device
        device = self.transformer.device

        if encoded_audio is not None:
            encoded_audio = encoded_audio.to(device)
            x_a = self.encoder_adapter(encoded_audio)
            if attention_mask_a is None:
                attention_mask_a = torch.ones(x_a.shape[0], x_a.shape[1], dtype=torch.long)
            attention_mask_a = attention_mask_a.to(device)
        else:
            x_a = torch.tensor([], device=device)
            attention_mask_a = torch.tensor([], device=device)

        if input_ids is not None:
            input_ids = input_ids.to(device)
            if input_ids.dim() == 1:
                input_ids = torch.unsqueeze(input_ids, 0)
            x_t = self.transformer.get_input_embeddings()(input_ids)
            if attention_mask_t is None:
                attention_mask_t = torch.ones(x_t.shape[0], x_t.shape[1], dtype=torch.long)
            attention_mask_t = attention_mask_t.to(device)
        else:
            x_t = torch.tensor([], device=device)
            attention_mask_t = torch.tensor([], device=device)

        x = torch.cat([x_a, x_t], dim=1)
        attention_mask = torch.cat([attention_mask_a, attention_mask_t], dim=1)

        output = self.transformer(
            inputs_embeds=x,
            attention_mask=attention_mask,
            past_key_values=self.past_key_values if enable_kvcache else None,
            labels=labels,
            num_logits_to_keep=labels.shape[1] if labels is not None else 0,
        )

        if enable_kvcache:
            self.past_key_values = output.past_key_values

        return output


# Copied from https://github.com/Lightning-AI/litgpt/blob/f6031e3a88e272ec86ad8f412573699589f4d41b/litgpt/generate/base.py#L30
def multinomial_num_samples_1(probs: torch.Tensor) -> torch.Tensor:
    if torch._dynamo.is_compiling():
        # Faster alternative to `torch.multinomial(probs, num_samples=1)` that is also CUDAGraph friendly
        distribution = torch.empty_like(probs).exponential_(1)
        return torch.argmax(probs / distribution, dim=-1, keepdim=True)
    return torch.multinomial(probs, num_samples=1)


def sample_top_p(logits: torch.Tensor, top_p: float) -> torch.Tensor:
    sorted_logits, sorted_indices = torch.sort(logits, descending=False)
    cumulative_probs = sorted_logits.softmax(dim=-1).cumsum(dim=-1)
    # Example:
    # sorted_probs=[0.1, 0.15, 0.2, 0.25, 0.3] -> sorted_cumprobs=[0.1, 0.25, 0.45, 0.7, 1.0]
    # sorted_indices_to_remove = [1, 1, 0, 0, 0] if top_p=0.7
    sorted_indices_to_remove = cumulative_probs <= (1 - top_p)
    # Keep at least 1 token always to prevent the case where no token is selected
    # In this case the most probable one is always kept
    sorted_indices_to_remove[-1:] = 0
    indices_to_remove = sorted_indices_to_remove.scatter(0, sorted_indices, sorted_indices_to_remove)
    logits = logits.masked_fill(indices_to_remove, float("-inf"))
    return logits


def sample(
    logits: torch.Tensor, temperature: float = 1.0, top_k: Optional[int] = None, top_p: float = 1.0
) -> torch.Tensor:
    if top_p < 0.0 or top_p > 1.0:
        raise ValueError(f"top_p must be in [0, 1], got {top_p}")
    logits = logits[0, -1]
    # optionally crop the logits to only the top k options
    if top_k is not None:
        v, i = torch.topk(logits, min(top_k, logits.size(-1)))
        # do not use `torch.where` as in nanogpt because it will repeat top-k collisions
        logits = torch.full_like(logits, float("-inf")).scatter_(-1, i, v)
    # optionally scale the logits and sample from a probability distribution
    if temperature > 0.0 or top_p > 0.0:
        if temperature > 0.0:
            logits = logits / temperature
        # optionally crop the logits to smallest set of logits with a cumulative probability above top_p
        if top_p < 1.0:
            logits = sample_top_p(logits, top_p)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        return multinomial_num_samples_1(probs)
    return torch.argmax(logits, dim=-1, keepdim=True)


## End of copied code


class AsrModelForGeneration(nn.Module):
    def __init__(self, in_features, **_):
        super().__init__()

        self.decoder = Decoder(audio_features_dim=in_features)
        self.decoder.freeze_lm()
        self.tokenizer = self.decoder.tokenizer
        self.sep_token = "<|vision_end|>"

        self.criterion = lambda pred, ref: torch.tensor(wer(ref, pred))

    def forward(self, batch_a: torch.Tensor, batch_t: torch.Tensor = None, return_loss: bool = True) -> torch.Tensor:
        result = self.decoder(
            encoded_audio=batch_a,
            attention_mask_a=None,
            input_ids=batch_t,
            attention_mask_t=None,
            labels=batch_t if return_loss else None,
        )
        if return_loss:
            return result.loss
        else:
            # Generate text for each audio input in the batch
            generated_texts = []
            for i in range(batch_a.shape[0]):
                single_audio = batch_a[i : i + 1]  # Keep batch dimension
                output_text = self.tokenizer.decode(self.generate(single_audio))
                generated_texts.append(output_text)

            label_texts = []
            assert batch_a.shape[0] == batch_t.shape[0]
            for i in range(batch_t.shape[0]):
                label_text = self.tokenizer.decode(batch_t[i])
                label_text = label_text.replace(self.sep_token, "").strip(self.tokenizer.pad_token)
                label_texts.append(label_text)

            return generated_texts, label_texts

    @torch.inference_mode()
    def generate(
        self,
        audio_features: torch.Tensor,
        max_returned_tokens: int = 200,
        *,
        temperature: float = 1.0,
        top_k: Optional[int] = None,
        top_p: float = 1.0,
    ) -> torch.Tensor:
        input_ids = self.decoder.tokenizer(self.sep_token, return_tensors="pt")
        self.decoder.clear_kvcache()
        logits = self.decoder(
            encoded_audio=audio_features,
            attention_mask_a=None,
            input_ids=input_ids["input_ids"],
            enable_kvcache=True,
        ).logits

        outputs = []
        next_t = sample(logits, temperature=temperature, top_k=top_k, top_p=top_p)

        while next_t != self.decoder.tokenizer.eos_token_id and len(outputs) < max_returned_tokens:
            outputs.append(next_t.item())
            logits = self.decoder(input_ids=next_t, enable_kvcache=True).logits
            next_t = sample(logits, temperature=temperature, top_k=top_k, top_p=top_p)
        return outputs



import os

import torch
import torch.nn as nn

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class Mlp(nn.Module):
    def __init__(self, in_features, out_features: int | None = None, criterion="CrossEntropyLoss"):
        super().__init__()
        out_features = out_features or in_features
        self.ln = nn.LayerNorm(in_features)
        self.fc = nn.Linear(in_features, out_features)
        self.criterion = getattr(nn, criterion)()

    def forward(self, x: torch.Tensor, y: torch.Tensor | None = None, return_loss: bool = False):
        x = self.ln(x)
        x = self.fc(x)
        if y is not None and return_loss:
            return self.criterion(x, y)
        return x, y



from xares.task import TaskConfig


def gtzan_genre_config(encoder) -> TaskConfig:
    class_label_maps = {
        "blues": 0,
        "classical": 1,
        "country": 2,
        "disco": 3,
        "hiphop": 4,
        "jazz": 5,
        "metal": 6,
        "pop": 7,
        "reggae": 8,
        "rock": 9,
    }
    config = TaskConfig(
        encoder=encoder,
        k_fold_splits=list(range(0, 10)),
        label_processor=lambda x: class_label_maps[x["genre"]],
        name="gtzan_genre",
        output_dim=len(class_label_maps),
        zenodo_id="14722472",
    )

    config.audio_tar_name_of_split = {fold: f"gtzan_fold_{fold}_0000000.tar" for fold in config.k_fold_splits}
    config.encoded_tar_name_of_split = {fold: f"gtzan-wds-encoded-fold-{fold}-*.tar" for fold in config.k_fold_splits}

    return config



from xares.task import TaskConfig


def libricount_config(encoder) -> TaskConfig:
    task = "num_speakers"
    class_label_maps = {i: i for i in range(11)}

    config = TaskConfig(
        batch_size_train=64,
        encoder=encoder,
        k_fold_splits=list(range(0, 5)),
        label_processor=lambda x: class_label_maps[x[task]],
        learning_rate=1e-3,
        name="libricount",
        output_dim=len(class_label_maps),
        zenodo_id="14722478",
    )

    config.audio_tar_name_of_split = {fold: f"libricount_fold{fold:02d}.tar" for fold in config.k_fold_splits}
    config.encoded_tar_name_of_split = {
        fold: f"libricount-wds-encoded-fold-{fold}-*.tar" for fold in config.k_fold_splits
    }

    return config



import argparse
from functools import partial

import pandas as pd
import torch
import torch.multiprocessing as mp
from loguru import logger

from xares.audio_encoder_checker import check_audio_encoder
from xares.common import setup_global_logger
from xares.metrics import weighted_average
from xares.task import XaresTask
from xares.utils import attr_from_py_path


def worker(
    encoder_py: None | str,
    task_py: str,
    do_download: bool = False,
    do_encode: bool = False,
    do_mlp: bool = False,
    do_knn: bool = False,
):
    # Encoder setup
    encoder = attr_from_py_path(encoder_py, endswith="Encoder")() if encoder_py else None

    # Task setup
    config = attr_from_py_path(task_py, endswith="_config")(encoder)
    if config.disabled:
        logger.warning(f"Task {config.name} is disabled, skipping")
        return (0, 0), (0, 0)
    task = XaresTask(config=config)

    # Run the task
    if do_download:
        logger.info(f"Downloading data for task {config.name} ...")
        task.download_audio_tar()
        logger.info(f"Task {config.name} data ready.")

    if do_encode:
        logger.info(f"Running make_encoded_tar for task {config.name} ...")
        task.make_encoded_tar()
        logger.info(f"Task {config.name} encoded.")

    if config.private and not (task.encoded_tar_dir / task.config.xares_settings.encoded_ready_filename).exists():
        logger.warning(f"Task {config.name} is private and not ready, skipping.")
        do_mlp = do_knn = False

    mlp_score = (0, 0)
    if do_mlp:
        logger.info(f"Running run_mlp for task {config.name} ...")
        mlp_score = task.run_mlp()
        logger.info(f"MLP score of {config.name}: {mlp_score}")

    knn_score = (0, 0)
    if do_knn and task.config.do_knn:
        logger.info(f"Running KNN for task {config.name} ...")
        knn_score = task.run_knn()
        logger.info(f"KNN score of {config.name}: {knn_score}")

    torch.cuda.empty_cache()
    return mlp_score, knn_score


def stage_1(encoder_py, task_py, gpu_id):
    torch.cuda.set_device(gpu_id)
    return worker(encoder_py, task_py, do_encode=True)


def stage_2(encoder_py, task_py, result: dict):
    result.update({task_py: worker(encoder_py, task_py, do_mlp=True, do_knn=True)})


def main(args):
    setup_global_logger()
    enable_multiprocessing = args.max_jobs > 0
    torch.multiprocessing.set_start_method("spawn")

    # Stage 0: Download all datasets
    stage_0 = partial(worker, do_download=True)
    if args.from_stage <= 0:
        try:
            if enable_multiprocessing:
                with mp.Pool(processes=args.max_jobs) as pool:
                    pool.starmap(stage_0, [(None, task_py) for task_py in args.tasks_py])
            else:
                for task_py in args.tasks_py:
                    stage_0(None, task_py)
            logger.info("Stage 0 completed: All data downloaded.")
        except Exception as e:
            if "Max retries exceeded with url" in str(e):
                logger.error(e)
                logger.error("This may be caused by Zenodo temporarily banning your connection.")
                logger.error("You may need to wait for a few hours and retry.")
                logger.error("Alternatively, you can download manually using `tools/download_manually.sh`.")
                return
            else:
                logger.error(f"Error in stage 0 (download): {e} Must fix it before proceeding.")
                return
    else:
        # Ensure pretrained model has been saved at local if stage 0 is skipped
        for task_py in args.tasks_py:
            worker(None, task_py)

    if args.to_stage == 0:
        return

    # Check if the encoder supports the multiprocessing
    if enable_multiprocessing:
        try:
            with mp.Pool(processes=1) as pool:
                pool.starmap(worker, [(args.encoder_py, args.tasks_py[0])])
        except Exception as e:
            logger.warning("Multiprocessing is not supported for the encoder. Falling back to a single process.")
            logger.warning("If single processing is too slow, you can manually parallelize tasks with a shell script.")
            logger.warning("For models from Hugging Face, try save locally, which might fix for multiprocessing.")
            enable_multiprocessing = False

    # Double check the encoder and download the pretrained weights
    encoder = attr_from_py_path(args.encoder_py, endswith="Encoder")()
    if not check_audio_encoder(encoder):
        raise ValueError("Invalid encoder")
    del encoder

    # Stage 1: Execute make_encoded_tar
    if args.from_stage <= 1:
        try:
            if enable_multiprocessing:
                num_gpus = torch.cuda.device_count()
                with mp.Pool(processes=args.max_jobs) as pool:
                    pool.starmap(
                        stage_1,
                        [(args.encoder_py, task_py, i % num_gpus) for i, task_py in enumerate(args.tasks_py)],
                    )
            else:
                for task_py in args.tasks_py:
                    worker(args.encoder_py, task_py, do_encode=True)

            logger.info("Stage 1 completed: All tasks encoded.")
        except FileNotFoundError as e:
            logger.error(f"Task filename pattern error: {e}")
            return
        except RuntimeError as e:
            if "CUDA out of memory" in str(e):
                logger.error("CUDA out of memory. Try reducing `config.batch_size_encode` of tasks.")
            else:
                logger.error(f"Error in stage 1 (encode): {e} Must fix it before proceeding.")
                return
        logger.info("Stage 1 completed: All tasks encoded.")
    if args.to_stage == 1:
        return

    # Stage 2: Execute mlp and knn scoring
    if args.from_stage <= 2 and args.to_stage >= 2:
        try:
            if enable_multiprocessing:
                manager = mp.Manager()
                return_dict = manager.dict()
                with mp.Pool(processes=args.max_jobs) as pool:
                    pool.starmap(
                        partial(stage_2, result=return_dict),
                        [(args.encoder_py, task_py) for task_py in args.tasks_py],
                    )
            else:
                return_dict = {}
                for task_py in args.tasks_py:
                    return_dict[task_py] = worker(args.encoder_py, task_py, do_mlp=True, do_knn=True)
        except RuntimeError as e:
            logger.error(f"Error in stage 2 (scoring): {e} Must fix it before proceeding.")
            return
        logger.info("Scoring completed: All tasks scored.")

        # Print results
        df = pd.DataFrame(return_dict.items(), columns=["Task", "Scores"])
        df["MLP_Score"] = df["Scores"].apply(lambda x: x[0][0])
        df["KNN_Score"] = df["Scores"].apply(lambda x: x[1][0])
        df["Task"] = df["Task"].str.replace(r".*/|_task\.py$", "", regex=True)
        df.drop(columns=["Scores"], inplace=True)
        df.sort_values(by="Task", inplace=True)

        print(f"\nResults:\n{df.to_string(index=False)}")

        avg_mlp, avg_knn = weighted_average(return_dict)
        print("\nWeighted Average MLP Score:", avg_mlp)
        print("Weighted Average KNN Score:", avg_knn)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run a task")
    parser.add_argument("encoder_py", type=str, help="Encoder path. eg: example/dasheng/dasheng_encoder.py")
    parser.add_argument(
        "tasks_py",
        type=str,
        help="Tasks path. eg: src/tasks/*.py",
        nargs="+",
    )
    parser.add_argument("--max-jobs", type=int, default=1, help="Maximum number of concurrent tasks.")
    parser.add_argument("--from-stage", default=0, type=int, help="First stage to run.")
    parser.add_argument("--to-stage", default=2, type=int, help="Last stage to run.")
    args = parser.parse_args()
    main(args)



import os

import torch
import torch.nn as nn
import torch.nn.functional as F
from loguru import logger

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class AudioTextContrastiveLoss(nn.Module):
    def __init__(self, temperature: float = 0.07, embed_regularization: bool = True):
        super().__init__()
        self.temperature = torch.nn.Parameter(torch.tensor(1.0 / temperature).log(), requires_grad=True)
        self.embed_regularization = embed_regularization

    def similarity(self, emb_x, emb_y):
        with torch.autocast(device_type="cuda", enabled=False):
            return self.temperature.exp() * emb_x @ emb_y.t()

    def forward(self, audio_embed, text_embed):
        sim_a2t = self.similarity(audio_embed, text_embed)
        sim_t2a = self.similarity(text_embed, audio_embed)

        sim_targets = torch.zeros(sim_a2t.size()).to(sim_a2t.device)
        sim_targets.fill_diagonal_(1)

        loss_a2t = -torch.sum(F.log_softmax(sim_a2t, dim=1) * sim_targets, dim=1).mean()

        loss_t2a = -torch.sum(F.log_softmax(sim_t2a, dim=1) * sim_targets, dim=1).mean()

        loss_atc = (loss_a2t + loss_t2a) / 2
        if self.embed_regularization:
            loss_atc = (
                loss_atc
                + torch.mean(torch.abs(audio_embed)) / torch.sqrt(torch.sum(audio_embed**2))
                + torch.mean(torch.abs(text_embed)) / torch.sqrt(torch.sum(text_embed**2))
            )
        return loss_atc


class RetrivalMLP(nn.Module):
    def __init__(
        self,
        in_features,
        hidden_features: int = 1024,
        out_features: int | None = None,
        criterion="AudioTextContrastiveLoss",
    ):
        super().__init__()
        out_features = out_features or hidden_features
        from transformers import AutoTokenizer

        self.ln = nn.LayerNorm(in_features)
        self.audio_mlp = nn.Sequential(
            nn.Linear(in_features, hidden_features), nn.GELU(), nn.Linear(hidden_features, out_features)
        )
        self.tokenizer = AutoTokenizer.from_pretrained("google-bert/bert-base-uncased")
        self.text_model = torch.nn.Embedding(self.tokenizer.vocab_size, hidden_features)
        self.pad_token_id = self.tokenizer.pad_token_id

        self.text_mlp = nn.Sequential(
            nn.Linear(hidden_features, hidden_features),
            nn.GELU(),
            nn.Linear(hidden_features, out_features),
        )

        self.criterion = globals()[criterion]()

    def forward(self, x: torch.Tensor, text: torch.Tensor, return_loss: bool = False):
        text_inputs = self.tokenizer(text, return_tensors="pt", padding=True)
        model_inputs = text_inputs.to(x.device)
        text_embeddings = self.text_model(model_inputs.input_ids)
        text_embeddings = self.text_mlp(text_embeddings)
        outputmask = (model_inputs.input_ids != self.pad_token_id).unsqueeze(-1).float()
        text_embeddings = (text_embeddings * outputmask).sum(1) / outputmask.sum(1)
        x = self.ln(x)
        x = self.audio_mlp(x)
        if return_loss:
            return self.criterion(x, text_embeddings)
        return x, text_embeddings



from loguru import logger

from xares.task import TaskConfig


def librispeech_asr_config(encoder) -> TaskConfig:
    config = TaskConfig(
        encoder=encoder,
        batch_size_train=1,
        disabled=True,
        do_knn=False,
        gradient_accumulation_steps=32,
        label_processor=None,
        metric="WER_inv",
        name="librispeech",
        pretrained_dependencies=["qwen2"],
        task_type="asr",
        test_split="test-clean",
        train_split="train-clean-100",
        use_mini_dataset=False,
        valid_split="dev-clean",
        zenodo_id="14716252",
    )

    if config.use_mini_dataset:
        logger.warning(f"Dataset {config.name} uses mini version for faster evaluation.")
        config.audio_tar_name_of_split[config.train_split] = "train-clean-100-000000.tar"

    return config



from xares.task import TaskConfig


def urbansound8k_config(encoder) -> TaskConfig:
    class_label_maps = {
        "air_conditioner": 0,
        "car_horn": 1,
        "children_playing": 2,
        "dog_bark": 3,
        "drilling": 4,
        "engine_idling": 5,
        "gun_shot": 6,
        "jackhammer": 7,
        "siren": 8,
        "street_music": 9,
    }
    config = TaskConfig(
        encoder=encoder,
        k_fold_splits=list(range(1, 11)),
        label_processor=lambda x: class_label_maps[x["soundevent"]],
        name="urbansound8k",
        output_dim=len(class_label_maps),
        zenodo_id="14722683",
    )
    config.audio_tar_name_of_split = {fold: f"urbansound_fold{fold}_0000000.tar" for fold in config.k_fold_splits}
    config.encoded_tar_name_of_split = {
        fold: f"urbansound-wds-encoded-fold-{fold}-*.tar" for fold in config.k_fold_splits
    }
    return config



from loguru import logger

from xares.task import TaskConfig


def nysnthinstument_config(encoder) -> TaskConfig:
    data_key = "instrument_family_str"
    class_label_maps = {
        "guitar": 0,
        "bass": 1,
        "organ": 2,
        "keyboard": 3,
        "vocal": 4,
        "string": 5,
        "reed": 6,
        "flute": 7,
        "mallet": 8,
        "brass": 9,
        "synth_lead": 10,
    }
    config = TaskConfig(
        encoder=encoder,
        label_processor=lambda x: class_label_maps[x[data_key]],
        name="nsynthinstument",
        output_dim=len(class_label_maps),
        test_split="nsynth_test",
        train_split="nsynth_train",
        valid_split="nsynth_valid",
        zenodo_id="14725174",
    )

    if config.use_mini_dataset:
        logger.warning(f"Dataset {config.name} uses mini version for faster evaluation.")
        config.audio_tar_name_of_split.update(
            {
                config.train_split: "nsynth_train_sub55k_0000000.tar",
                config.valid_split: "nsynth_valid_0000001.tar",
            }
        )

    return config



from __future__ import annotations

import math
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from typing import Callable, Dict, List, Literal, Tuple

import numpy as np
import torch
from ignite.metrics import Accuracy, AveragePrecision, EpochMetric, MeanAbsoluteError, MeanSquaredError, Metric

METRICS_TYPE = Literal["accuracy", "EER", "mAP", "recallatk_r1", "MAE", "MSE", "segmentf1", "WER_inv"]


@dataclass
class EvalMetric:
    metric: Callable
    score: float = 1.0  # can also be -1


class ClapScore(Metric):
    def __init__(
        self,
        num_caps: int = 5,  # Clotho uses 5 caps
        select: Literal["r1", "r5", "r10", "mAP10"] | None = None,
        average: bool = True,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._reset()
        self.average = average
        self.num_caps = num_caps
        self.select = select

    def _reset(self):
        self.embds_x, self.embds_y = [], []

    def update(self, output):
        """update.

        :param self: self object
        """
        embeds_x, embeds_y = output
        self.embds_x.append(embeds_x)
        self.embds_y.append(embeds_y)

    def reset(self):
        self._reset()

    def compute(self) -> Dict[str, float] | float:
        return _clapscore(
            torch.cat(self.embds_x, dim=0),
            torch.cat(self.embds_y, dim=0),
            num_caps=self.num_caps,
            average=self.average,
            select=self.select,
        )


class SegmentF1Metric(Metric):
    def __init__(self, *args, output_transform=lambda x: x, hop_size_in_ms: float, segment_length_in_s: float = 1.0):
        super().__init__(*args, output_transform=output_transform)
        self._reset()
        self.max_pooling_size = math.ceil(segment_length_in_s / (hop_size_in_ms / 1000.0))

    def _reset(self):
        self.pred, self.targets = [], []

    def update(self, output):
        pred, tar = output
        assert pred.ndim == tar.ndim, "Dims need to be identical"
        pred = (
            torch.nn.functional.max_pool1d(pred.transpose(-2, -1), kernel_size=self.max_pooling_size)
            .transpose(-2, -1)
            .flatten(0, 1)
        )
        tar = (
            torch.nn.functional.max_pool1d(tar.transpose(-2, -1), kernel_size=self.max_pooling_size)
            .transpose(-2, -1)
            .flatten(0, 1)
        )
        self.pred.append(pred)
        self.targets.append(tar)

    def reset(self):
        self._reset()

    def compute(self) -> Dict[str, float] | float:
        from sklearn.metrics import f1_score

        # Binary classification, preds need to be {0,1} and targets
        pred = torch.cat(self.pred).long().cpu().numpy()
        tar = torch.cat(self.targets).long().cpu().numpy()
        return f1_score(tar, pred, average="macro")


def compute_eer(pred, target, positive_label: int = 1):
    from sklearn.metrics import roc_curve

    fpr, tpr, _ = roc_curve(target, pred, positive_label)
    fnr = 1 - tpr
    eer_1 = fpr[np.nanargmin(np.absolute((fnr - fpr)))]
    eer_2 = fnr[np.nanargmin(np.absolute((fnr - fpr)))]
    eer = (eer_1 + eer_2) / 2
    return eer


EERMetric = lambda: EpochMetric(compute_fn=compute_eer)


class WerScore(Metric):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._reset()

    def _reset(self):
        self.pred, self.target = [], []

    def update(self, output):
        pred, target = output
        self.pred.extend(pred)
        self.target.extend(target)

    def reset(self):
        self._reset()

    def compute(self) -> Dict[str, float] | float:
        from jiwer import wer

        wer_score = max(0, 1 - wer(self.target, self.pred))
        return wer_score


ALL_METRICS: Dict[METRICS_TYPE, EvalMetric] = dict(
    accuracy=EvalMetric(Accuracy, score=1.0),
    frame_mAP=EvalMetric(
        partial(AveragePrecision, output_transform=lambda x: (torch.flatten(x[0], 0, 1), torch.flatten(x[1], 0, 1))),
        score=1.0,
    ),
    segmentf1=EvalMetric(
        partial(SegmentF1Metric, output_transform=lambda x: (x[0].sigmoid().round(), x[1])), score=1.0
    ),
    mAP=EvalMetric(AveragePrecision, score=1.0),
    MAE=EvalMetric(MeanAbsoluteError, score=-1.0),
    MSE=EvalMetric(MeanSquaredError, score=-1.0),
    recallatk_r1=EvalMetric(partial(ClapScore, select="r1"), score=1.0),
    EER=EvalMetric(EERMetric, score=-1.0),
    WER_inv=EvalMetric(WerScore, score=1.0),
)


def cos_sim(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    a = torch.nn.functional.normalize(a, p=2, dim=1)
    b = torch.nn.functional.normalize(b, p=2, dim=1)
    return a @ b.transpose(0, 1)


def clap_metrics(
    audio_embds: torch.Tensor,  # size of (B, D)
    text_embds: torch.Tensor,  # size of (B, D)
    num_caps: int = 5,
    topk: int = 10,
    topk_recall: List[int] = [1, 5, 10],
):
    metrics = {}
    n_embds = len(audio_embds)
    logits_per_audio = cos_sim(audio_embds, text_embds).detach()
    logits_per_text = logits_per_audio.t().detach()

    logits = {"a2t": logits_per_audio, "t2a": logits_per_text}

    for name, logit in logits.items():
        # take the first sample when we have num_caps embeddings
        if name == "a2t":
            ground_truth = torch.arange(n_embds, device=audio_embds.device).view(-1, 1)
            logit = logit.view(n_embds // num_caps, num_caps, -1)[:, 0, :].repeat_interleave(num_caps, dim=0)
        else:
            ground_truth = (
                torch.arange(n_embds // num_caps, device=audio_embds.device).repeat_interleave(num_caps).view(-1, 1)
            )
            logit = logit.view(-1, n_embds // num_caps, num_caps)[:, :, 0]

        ranking = torch.argsort(logit, descending=True)
        preds = torch.where(ranking == ground_truth)[1]
        # Calcualte map@10
        ranks = np.sort(preds.view(-1, num_caps).cpu().numpy() + 1, axis=-1)
        if name == "a2t":
            ap = np.arange(1, num_caps + 1) / ranks
        else:
            ap = 1.0 / ranks
        ap[ranks > topk] = 0
        # Per caps
        ap = np.mean(ap, axis=-1)
        mAP10 = 100 * np.mean(ap)

        # Minimum rank for each sample
        if name == "a2t":
            preds = preds.view(-1, num_caps).min(-1)[0]

        preds = preds.detach().cpu().numpy()
        metric_dict = {}
        for k in topk_recall:
            metric_dict[f"r{k}"] = np.mean(preds < k) * 100.0
        metric_dict["meanr"] = preds.mean() + 1
        metric_dict["medr"] = np.floor(np.median(preds)) + 1
        metric_dict["mAP10"] = mAP10
        metrics[name] = metric_dict
    return metrics


def _clapscore(
    *args,
    num_caps: int = 1,
    select: Literal["r1", "r5", "r10", "mAP10"] | None = None,
    average: bool = True,
) -> Dict[str, float] | float:
    ret = clap_metrics(*args, num_caps=num_caps)
    if average:
        average_dict = defaultdict(list)
        for metric, metric_dict in ret.items():
            for k, v in metric_dict.items():
                average_dict[k].append(v)
        ret = {k: np.mean(v) for k, v in average_dict.items()}
    else:
        result = {}
        for metric, metric_dict in ret.items():
            for k, v in metric_dict.items():
                result[f"{metric}_{k}"] = v
        ret = result
    if select:
        ret = ret[select]
    return ret / 100


def weighted_average(scores_dict: Dict[str, List[Tuple[float, int]]]) -> List[float]:
    """
    Compute the weighted average of scores.

    Args:
        scores_dict (Dict[str, List[Tuple[float, int]]]):
            A dictionary where the key is a dataset name and the value is a list of tuples.
            Each tuple contains a score and its corresponding weight.

    Returns:
        List[float]: A list of weighted average scores,
        where each element is the weighted average for scores at the same index across datasets.

    Example:
        scores_dict = {
            'dataset_1': [(0.8, 100), (0.6, 50)],
            'dataset_2': [(0.9, 200), (0.7, 100)]
        }
        result = weighted_average(scores_dict)
        # result should be approximately: [
        #   (0.8 * 100 + 0.9 * 200) / (100 + 200),  # Weighted average for the first score index across datasets
        #   (0.6 * 50 + 0.7 * 100) / (50 + 100)   # Weighted average for the second score index across datasets
        # ]
        # result should be approximately [0.8666666666666667, 0.6666666666666666]
    """

    output = []
    num_score_types = len(list(scores_dict.values())[0]) if scores_dict else 0  # Handle empty scores_dict
    for score_index in range(num_score_types):
        weighted_sum = 0
        total_weight = 0
        for dataset_scores in scores_dict.values():
            if score_index < len(dataset_scores):  # Check if score_index is valid for current dataset
                score, weight = dataset_scores[score_index]
                weighted_sum += score * weight
                total_weight += weight
        weighted_avg = weighted_sum / total_weight if total_weight != 0 else 0
        output.append(weighted_avg)
    return output



from loguru import logger

from xares.task import TaskConfig


def asvspoof2015_config(encoder) -> TaskConfig:
    data_key = "binary_spoof"
    class_label_maps = {"human": 0, "spoof": 1}
    config = TaskConfig(
        batch_size_train=64,
        encoder=encoder,
        epochs=5,
        label_processor=lambda x: class_label_maps[x[data_key]],
        learning_rate=1e-3,
        name="asvspoof2015",
        output_dim=len(class_label_maps),
        test_split="asvspoof_eval",
        train_split="asvspoof_train",
        valid_split="asvspoof_valid",
        zenodo_id="14718430",
    )

    if config.use_mini_dataset:
        logger.warning(f"Dataset {config.name} uses mini version for faster evaluation.")
        config.audio_tar_name_of_split = {
            config.train_split: "asvspoof_train_0000000.tar",
            config.valid_split: "asvspoof_valid_0000000.tar",
            config.test_split: "asvspoof_eval_0000000.tar",
        }

    return config



from __future__ import annotations

import io
import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Tuple

import numpy as np
import torch
import torch.multiprocessing as mp
import torch.nn as nn
import webdataset as wds
from loguru import logger
from tqdm import tqdm

from xares.audiowebdataset import create_embedding_webdataset, create_rawaudio_webdataset
from xares.common import XaresSettings
from xares.metrics import METRICS_TYPE
from xares.models.asr import AsrModelForGeneration
from xares.models.retrival import RetrivalMLP
from xares.models.simple import Mlp
from xares.trainer import KNNTrainer, Trainer
from xares.utils import download_hf_model_to_local, download_zenodo_record, mkdir_if_not_exists


@dataclass
class TaskConfig:
    name: str
    xares_settings: XaresSettings = field(default_factory=XaresSettings)
    env_root: Path | str | None = None
    disabled: bool = False  # Skip the task, useful for debugging

    # General
    private: bool = False
    torch_num_threads: int = 2  # Do not use too many otherwise slows down
    seed: int = 42  # manual seed for all experiments
    label_processor: Callable | None = None
    merge_processor: Callable | None = None
    task_type: Literal["frame", "clip", "contrastive"] = "clip"

    # Splits
    train_split: None | str = "train"
    valid_split: None | str = "valid"
    test_split: None | str = "test"
    k_fold_splits: None | List[int | str] = None
    use_mini_dataset: bool = True  # For some large datasets, use subset for faster evaluation

    # Audio tar
    force_download: bool = False
    audio_tar_name_of_split: Dict[Any, Any] = field(default_factory=lambda: dict())
    num_shards_rawaudio = 4
    zenodo_id: str | None = None

    # Encoded tar
    force_encode: bool = False
    encoder: None | Any = None
    pretrained_dependencies: None | List[str] = None
    encoded_tar_name_of_split: Dict[Any, Any] = field(default_factory=lambda: dict())
    trim_length = None
    save_encoded_per_batches: int = 2000
    batch_size_encode: int = 16
    num_encoder_workers: int = 4
    crop_length: None | float = None

    # MLP
    force_retrain_mlp: bool = False
    ckpt_dir_name = "checkpoints"
    embedding_dir_name = "embeddings"
    ckpt_name = "best.ckpt"
    criterion: Literal["CrossEntropyLoss", "BCEWithLogitsLoss", "L1Loss", "MSELoss", "AudioTextContrastiveLoss"] = (
        "CrossEntropyLoss"
    )
    batch_size_train: int = 32
    gradient_accumulation_steps: int = 1
    learning_rate: float = 1e-3
    epochs: int = 10
    num_training_workers: int = 0
    num_validation_workers: int = 0
    model: nn.Module | None = None
    output_dim: int | None = None
    metric: METRICS_TYPE = "accuracy"
    metric_args: Dict[str, Any] = field(default_factory=lambda: dict())

    # KNN
    do_knn: bool = True

    def __post_init__(self, **kwargs):
        self.update_tar_name_of_split()
        if self.env_root is None:
            self.env_root = self.xares_settings.env_root
        torch.set_num_threads(self.torch_num_threads)
        torch.manual_seed(self.seed)
        np.random.seed(self.seed)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def update_tar_name_of_split(self):
        if self.k_fold_splits is not None:
            self.audio_tar_name_of_split = {fold: f"wds-audio-fold-{fold}-*.tar" for fold in self.k_fold_splits}
            self.encoded_tar_name_of_split = {fold: f"wds-encoded-fold-{fold}-*.tar" for fold in self.k_fold_splits}
        else:
            self.audio_tar_name_of_split = {
                self.train_split: f"{self.train_split}*.tar",
                self.valid_split: f"{self.valid_split}*.tar",
                self.test_split: f"{self.test_split}*.tar",
            }
            self.encoded_tar_name_of_split = {
                self.train_split: f"wds-encoded-{self.train_split}*.tar",
                self.valid_split: f"wds-encoded-{self.valid_split}*.tar",
                self.test_split: f"wds-encoded-{self.test_split}*.tar",
            }


class XaresTask:
    def __init__(self, config: TaskConfig):
        self.config = config

        if self.config.encoder is not None:
            self.encoder_device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            self.encoder = self.config.encoder.to(self.encoder_device)
            self.encoder_name = self.encoder.__class__.__name__
        else:
            self.encoder = None
            self.encoder_name = "Unknown"

        torch.set_num_threads(self.config.torch_num_threads)
        torch.manual_seed(self.config.seed)
        np.random.seed(self.config.seed)

        self.env_dir = Path(self.config.env_root) / self.config.name
        self.ckpt_dir = self.env_dir / self.config.ckpt_dir_name / self.encoder_name
        self.encoded_tar_dir = self.env_dir / self.config.embedding_dir_name / self.encoder_name
        self.encoded_ready_path = self.encoded_tar_dir / self.config.xares_settings.encoded_ready_filename
        self.ckpt_path = self.ckpt_dir / self.config.ckpt_name
        mkdir_if_not_exists(self.encoded_tar_dir)

        if self.config.k_fold_splits:
            self.config.train_split = self.config.valid_split = self.config.test_split = None

        self.label_processor = self.config.label_processor
        self.merge_processor = self.config.merge_processor

        if self.config.task_type == "contrastive":
            self.mlp_model = RetrivalMLP
        elif self.config.task_type == "asr":
            self.mlp_model = AsrModelForGeneration
        else:
            self.mlp_model = Mlp

        self.encoded_tar_path_of_split = {
            split: (self.encoded_tar_dir / self.config.encoded_tar_name_of_split[split])
            for split in self.config.encoded_tar_name_of_split
        }

        if self.config.pretrained_dependencies is not None:
            download_hf_model_to_local(self.config.pretrained_dependencies)

    def download_audio_tar(self):
        if self.config.private:
            logger.warning(f"Dataset {self.config.name} is private. Do not download from Zenodo.")
            return

        audio_ready_path = self.env_dir / self.config.xares_settings.audio_ready_filename
        if not self.config.force_download and audio_ready_path.exists():
            logger.warning(f"Skip downloading audio tar: {audio_ready_path} exists.")
            return

        download_zenodo_record(self.config.zenodo_id, self.env_dir, force_download=self.config.force_download)
        audio_ready_path.touch()

    def make_encoded_tar(self):
        if not self.config.force_encode and self.encoded_ready_path.exists():
            logger.warning(f"Skip encoding: {self.encoded_ready_path} exists.")
            return

        audio_ready_path = self.env_dir / self.config.xares_settings.audio_ready_filename
        if not audio_ready_path.exists():
            if self.config.private:
                logger.warning(f"For private dataset {self.config.name}, data must be placed at local path. Skip.")
                return
            download_zenodo_record(self.config.zenodo_id, self.env_dir, force_download=self.config.force_download)
            audio_ready_path.touch()

        audio_tar_path_of_split = {
            split: (self.env_dir / self.config.audio_tar_name_of_split[split]).as_posix()
            for split in self.config.audio_tar_name_of_split
        }

        for split in audio_tar_path_of_split:
            logger.info(f"Encoding audio for split {split} ...")
            logger.debug(f"Using data from {audio_tar_path_of_split[split]} ... ")
            dl = create_rawaudio_webdataset(
                [audio_tar_path_of_split[split]],
                target_sample_rate=self.encoder.sampling_rate,
                audio_key_name="audio",
                num_workers=(
                    0
                    if (mp.current_process().daemon or mp.get_start_method() != "fork")
                    else self.config.num_encoder_workers
                ),
                batch_size=self.config.batch_size_encode,
                crop_length=self.config.crop_length,
                pad_last=True,  # Add crop
            )
            sink = wds.ShardWriter(
                self.encoded_tar_path_of_split[split].as_posix().replace("*", f"0%05d"),
                encoder=False,
                compress=False,
                maxcount=self.config.save_encoded_per_batches,
                verbose=False,
            )

            with torch.inference_mode():
                for enum_item, ((audio, _), json_data, filenames) in tqdm(
                    enumerate(dl), desc=f"Encoding {split}", leave=True
                ):
                    audio = audio.to(self.encoder_device)
                    embedding = self.encoder(audio).to("cpu").detach()
                    for embed, json_data_sample, filename in zip(embedding, json_data, filenames):
                        buf = io.BytesIO()
                        np.save(buf, embed.numpy())
                        sink.write(
                            {
                                "npy": buf.getvalue(),
                                "json": json.dumps(json_data_sample).encode("utf-8"),
                                "__key__": f"{filename}{enum_item}",
                            }
                        )

        self.encoded_ready_path.touch()

    def run_mlp(self) -> Tuple[float, int]:
        mlp_score = 0
        eval_size = 0
        score_file = self.ckpt_dir / "mlp_score.txt"

        if score_file.exists():
            with open(score_file, "r") as f:
                lines = f.read().splitlines()
                mlp_score = float(lines[0])
                eval_size = int(lines[1])
            logger.info(f"Loaded MLP score from {score_file}: {mlp_score}")
            return mlp_score, eval_size

        if self.config.k_fold_splits:
            # K-fold cross validation
            acc = []
            eval_sizes = []
            splits = self._make_splits()
            wds_encoded_training_fold_k = {
                k: [self.encoded_tar_path_of_split[j].as_posix() for j in splits if j != k] for k in splits
            }

            for k in splits:
                self.config.ckpt_name = f"fold_{k}_best_model.pt"
                self.ckpt_path = self.ckpt_dir / self.config.ckpt_name
                self.train_mlp(
                    wds_encoded_training_fold_k[k],
                    [self.encoded_tar_path_of_split[k].as_posix()],
                )
                score, eval_size = self.evaluate_mlp([self.encoded_tar_path_of_split[k].as_posix()], load_ckpt=True)
                acc.append(score)
                eval_sizes.append(eval_size)

            for k in range(len(splits)):
                logger.info(f"Fold {k+1} {self.config.metric}: {acc[k]}")

            avg_score = np.mean(acc)
            total_eval_size = int(np.average(eval_sizes))
            logger.info(f"The averaged {self.config.metric} of 5 folds is: {avg_score}")

            mlp_score = avg_score
            eval_size = total_eval_size
        else:
            # Single split
            self.train_mlp(
                [self.encoded_tar_path_of_split[self.config.train_split].as_posix()],
                [self.encoded_tar_path_of_split[self.config.valid_split].as_posix()],
            )
            mlp_score, eval_size = self.evaluate_mlp(
                [self.encoded_tar_path_of_split[self.config.test_split].as_posix()],
                load_ckpt=True,
            )
            logger.info(f"The {self.config.metric}: {mlp_score}")

        with open(score_file, "w") as f:
            f.write(f"{mlp_score}\n{eval_size}")
        logger.info(f"Saved MLP score to {score_file}: {mlp_score}")

        return mlp_score, eval_size

    def train_mlp(self, train_url: list, validation_url: list) -> None:
        mlp = self.mlp_model(
            in_features=self.encoder.output_dim, out_features=self.config.output_dim, criterion=self.config.criterion
        )

        self.trainer = Trainer(
            mlp,
            ckpt_dir=self.ckpt_dir,
            ckpt_name=self.config.ckpt_name,
            metric=self.config.metric,
            metric_args=self.config.metric_args,
            gradient_accumulation_steps=self.config.gradient_accumulation_steps,
            lr=self.config.learning_rate,
            max_epochs=self.config.epochs,
            task_type=self.config.task_type,
        )

        if not self.config.force_retrain_mlp and self.ckpt_path.exists():
            logger.info(f"Checkpoint {self.ckpt_path} already exists. Skip training.")
            self.trainer.load_state_dict(torch.load(self.ckpt_path))
            return

        dl_train = create_embedding_webdataset(
            train_url,
            tar_shuffle=2000,
            batch_size=self.config.batch_size_train,
            num_workers=self.config.num_training_workers,
            training=True,
            label_processor=self.label_processor,
            merge_processor=self.merge_processor,
        )
        dl_val = create_embedding_webdataset(
            validation_url,
            batch_size=self.config.batch_size_train,
            num_workers=self.config.num_validation_workers,
            training=False,
            label_processor=self.label_processor,
            merge_processor=self.merge_processor,
        )

        try:
            self.trainer.run(dl_train, dl_val)
        except RuntimeError as e:
            if "at least one example" in str(e):
                raise RuntimeError(f"Empty dataloader. Try delete {self.encoded_ready_path} and re-run.")
            else:
                raise e

    def evaluate_mlp(self, eval_url: list, load_ckpt: bool = False) -> Tuple[Dict[METRICS_TYPE, Any], int]:
        if self.trainer is None:
            raise ValueError("Train the model first before evaluation.")

        if load_ckpt:
            if self.ckpt_path.exists():
                self.trainer.load_state_dict(torch.load(self.ckpt_path))
                logger.info(f"Loaded model parameters from {self.ckpt_path}")
            else:
                logger.warning(f"No checkpoint found at {self.ckpt_path}. Skip loading.")

        dl = create_embedding_webdataset(
            eval_url,
            batch_size=self.config.batch_size_train,
            num_workers=self.config.num_validation_workers,
            label_processor=self.label_processor,
            merge_processor=self.merge_processor,
            training=False,
        )
        return self.trainer.run_inference(dl)  # (result, size)

    def run_knn(self) -> Tuple[float, int]:
        knn_score = 0
        eval_size = 0
        if not self.config.do_knn:
            logger.warning(f"Skip KNN evaluation for {self.config.name}.")
            return knn_score, eval_size

        score_file = self.ckpt_dir / "knn_score.txt"

        if score_file.exists():
            with open(score_file, "r") as f:
                lines = f.read().splitlines()
                knn_score = float(lines[0])
                eval_size = int(lines[1])
            logger.info(f"Loaded KNN score from {score_file}: {knn_score}")
            return knn_score, eval_size

        if self.config.k_fold_splits:
            # K-fold cross validation
            scores = []
            eval_sizes = []
            splits = self._make_splits()
            wds_encoded_training_fold_k = {
                k: [self.encoded_tar_path_of_split[j].as_posix() for j in splits if j != k] for k in splits
            }

            for k in splits:
                score, size = self.train_and_eval_knn(
                    wds_encoded_training_fold_k[k],
                    [self.encoded_tar_path_of_split[k].as_posix()],
                )
                scores.append(score)
                eval_sizes.append(size)

            for k in range(len(splits)):
                logger.info(f"Fold {k+1} {self.config.metric}: {scores[k]}")

            avg_score = np.mean(scores)
            total_eval_size = int(np.average(eval_sizes))
            logger.info(f"The averaged KNN {self.config.metric} of 5 folds is: {avg_score}")

            knn_score = avg_score
            eval_size = total_eval_size
        else:
            # Single split
            score, size = self.train_and_eval_knn(
                [self.encoded_tar_path_of_split[self.config.train_split].as_posix()],
                [self.encoded_tar_path_of_split[self.config.test_split].as_posix()],
            )
            logger.info(f"The KNN score: {score}")
            knn_score = score
            eval_size = size

        with open(score_file, "w") as f:
            f.write(f"{knn_score}\n{eval_size}")
        logger.info(f"Saved KNN score to {score_file}: {knn_score}")

        return knn_score, eval_size

    def train_and_eval_knn(self, train_url, eval_url) -> Tuple[float, int]:
        dl_train = create_embedding_webdataset(
            train_url,
            tar_shuffle=2000,
            batch_size=self.config.batch_size_train,
            num_workers=self.config.num_training_workers,
            training=True,
            label_processor=self.label_processor,
        )
        dl_eval = create_embedding_webdataset(
            eval_url,
            batch_size=self.config.batch_size_train,
            num_workers=self.config.num_validation_workers,
            training=False,
            label_processor=self.label_processor,
        )
        knn_trainer = KNNTrainer(num_classes=self.config.output_dim)
        return knn_trainer.train_and_eval(dl_train, dl_eval)

    def run(self):
        self.download_audio_tar()
        self.make_encoded_tar()
        mlp_score = self.run_mlp()
        knn_score = self.run_knn()
        return mlp_score, knn_score

    def _make_splits(self):
        splits = self.config.k_fold_splits or [self.config.train_split, self.config.valid_split, self.config.test_split]
        return list(filter(lambda x: x is not None, splits))



import logging
import sys
from dataclasses import dataclass

from loguru import logger


def setup_global_logger():
    logging.captureWarnings(True)
    logging.getLogger("py.warnings").setLevel(logging.ERROR)

    # Make the logger with this format the default for all loggers in this package
    logger.configure(
        handlers=[
            {
                "sink": sys.stdout,
                "format": "<fg #FF6900>(X-ARES)</fg #FF6900> [<yellow>{time:YYYY-MM-DD HH:mm:ss}</yellow>] "
                "<level>{message}</level>",
                "level": "DEBUG",
                "colorize": True,
            }
        ]
    )
    logger.level("ERROR", color="<red>")
    logger.level("INFO", color="<white>")


@dataclass()
class XaresSettings:
    env_root: str = "./env"
    audio_ready_filename: str = ".audio_tar_ready"
    encoded_ready_filename: str = ".encoded_tar_ready"

    def __post_init__(self):
        setup_global_logger()



from xares.task import TaskConfig


def data_merge_function_clotho(data_stream):
    for sample in data_stream:
        audio, captions, *extra = sample
        for caption in captions.values():
            yield audio, caption, *extra


def clotho_config(encoder) -> TaskConfig:
    task_config = TaskConfig(
        batch_size_train=128,
        criterion="AudioTextContrastiveLoss",
        crop_length=30,
        do_knn=False,
        encoder=encoder,
        epochs=20,
        merge_processor=data_merge_function_clotho,
        metric="recallatk_r1",
        name="clotho",
        pretrained_dependencies=["bert-base-uncased:tokenizer"],
        save_encoded_per_batches=500,
        task_type="contrastive",
        test_split="clotho_validation",  # Worst naming scheme in a dataset
        train_split="clotho_development",
        valid_split="clotho_evaluation",
        zenodo_id="14856454",
    )
    return task_config



from xares.task import TaskConfig


def liveenv_config(encoder) -> TaskConfig:
    data_key = "soundevent"
    class_label_maps = {
        "Airplane": 0,
        "AirportHall": 1,
        "Alarm": 2,
        "BabyCry": 3,
        "BabyLaugh": 4,
        "BabySound ": 5,
        "Car": 6,
        "Cat": 7,
        "DiningArea": 8,
        "Dog": 9,
        "InsideTrain": 10,
        "Keyboard": 11,
        "KnockDoor": 12,
        "Metro": 13,
        "MetroWaitingArea": 14,
        "Motorbike": 15,
        "Ocean": 16,
        "Office": 17,
        "Park": 18,
        "Poultry": 19,
        "PublicSquareDance": 20,
        "Restaurant": 21,
        "Sheep": 22,
        "ShoppingMall": 23,
        "VaccumCleaner": 24,
        "WaitingRoom": 25,
        "Washmashine": 26,
        "Water": 27,
    }

    return TaskConfig(
        do_knn=False,
        encoder=encoder,
        epochs=50,
        label_processor=lambda x: class_label_maps[x[data_key]],
        name="liveenv",
        output_dim=len(class_label_maps),
        private=True,
        test_split="liveenv_test",
        train_split="liveenv_train",
        valid_split="liveenv_test",
    )



from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from functools import partial
from typing import Any, Dict, Iterable, Literal, Tuple

import torch
import torch.nn as nn
from ignite.engine import Engine, Events
from ignite.handlers import CosineAnnealingScheduler, EpochOutputStore, global_step_from_engine
from ignite.handlers.tqdm_logger import ProgressBar
from ignite.metrics import Loss, RunningAverage
from loguru import logger
from torch import Tensor, nn, optim
from torch.nn.functional import normalize

from xares.metrics import ALL_METRICS, METRICS_TYPE


def length_to_mask(length):
    max_length = length.amax().item()
    idx = torch.arange(max_length, device=length.device)
    return idx.unsqueeze(0) < length.unsqueeze(-1)


def masked_mean(x, x_length, dim: int = -1):
    mask = length_to_mask(x_length)
    mask_target_shape = (len(mask),) + (1,) * (x.ndim - 2) + (mask.shape[-1],)
    mask = mask.view(mask_target_shape)
    return (x * mask).sum(dim) / mask.sum(dim)


def cast_to_tensor(y: Iterable):
    y = y if isinstance(y, torch.Tensor) else torch.tensor(y)
    return y.float() if y.is_floating_point() else y


def pad_or_trim(tensor, target_size):
    difference = target_size - tensor.shape[-1]
    if difference < 0:
        return tensor[..., :target_size].contiguous()  # Trim excess values
    elif difference > 0:
        return torch.nn.functional.pad(tensor, pad=(0, difference))
    return tensor


def prepare_clip_task_batch(batch: Tuple, device: torch.device | str = "cpu"):
    (x, x_length), y, _ = batch
    x = masked_mean(x, x_length=x_length, dim=-1)
    y = cast_to_tensor(y)
    return x.to(device), y.to(device)


def prepare_contrastive_task_batch(batch: Tuple, device: torch.device | str = "cpu"):
    (x, x_length), y, _ = batch
    # y is a list of strings
    x = masked_mean(x, x_length=x_length, dim=-1)
    # x, y are (B,D,T), but models need B,T,D
    return x.to(device), y


def prepare_frame_task_batch(batch: Tuple, device: torch.device | str = "cpu"):
    (x, x_length), (y, y_length), _ = batch
    y = cast_to_tensor(y)
    # Trim the labels
    y = pad_or_trim(y, x.shape[-1])
    # x, y are (B,D,T), but models need B,T,D
    return x.transpose(-2, -1).contiguous().to(device), y.transpose(-2, -1).contiguous().to(device)


def prepare_asr_task_batch(
    batch: Tuple, device: torch.device | str = "cpu", tokenizer: Any = None, sep_token: str = "<|vision_end|>"
):
    if tokenizer is None:
        raise ValueError("Tokenizer must be provided for ASR task.")

    from transformers import DataCollatorForLanguageModeling

    (x, _), labels, _ = batch

    text_with_eos_bos = [f"{sep_token}{label['trans']}{tokenizer.eos_token}" for label in labels]
    tokens = [tokenizer(text) for text in text_with_eos_bos]
    data_collator_for_lm = DataCollatorForLanguageModeling(tokenizer, mlm=False)
    y = data_collator_for_lm(tokens)["input_ids"]

    # x are (B,D,T), but models need B,T,D
    return x.transpose(-2, -1).contiguous().to(device), y.contiguous().to(device)


@dataclass
class Trainer:
    model: nn.Module
    device: torch.device | Literal["cpu"] = field(
        default_factory=lambda: torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    )
    optimizer: str = "Adam"
    lr: float = 3e-3
    gradient_accumulation_steps: int = 1
    max_epochs: int = 10
    ckpt_dir: str = "checkpoints"
    best_ckpt_path: str | None = None
    ckpt_name: str = "best_model.pt"
    metric: METRICS_TYPE = "accuracy"
    metric_args: Dict[str, Any] = field(default_factory=lambda: dict())
    label_name: str = "target"
    save_model: bool = True
    decay_fraction: float = 0.1  # Decay learning rate
    task_type: InitVar[Literal["frame", "clip", "contrastive"]] = "clip"

    def __post_init__(self, task_type: Literal["frame", "clip", "contrastive"] = "clip"):
        try:
            self.optimizer = getattr(optim, self.optimizer)(self.model.parameters(), lr=self.lr)
        except AttributeError:
            raise NotImplementedError(f"Optimizer {self.optimizer} not implemented.")

        if task_type == "clip":
            self.prepare_batch_function = prepare_clip_task_batch
        elif task_type == "frame":
            self.prepare_batch_function = prepare_frame_task_batch
        elif task_type == "contrastive":
            self.prepare_batch_function = prepare_contrastive_task_batch
        elif task_type == "asr":
            self.prepare_batch_function = partial(prepare_asr_task_batch, tokenizer=self.model.tokenizer)
        else:
            raise NotImplementedError(f"Trainer.prepare_batch_function for task_type {task_type} not implemented.")

        self.ignite_trainer = Engine(self.train_step)
        self.ignite_evaluator = Engine(self.validation_step)
        # Schedule cosine annealing during training
        if self.max_epochs > 1:
            scheduler = CosineAnnealingScheduler(
                self.optimizer,
                "lr",
                self.optimizer.param_groups[0]["lr"],
                self.lr * self.decay_fraction,
                self.max_epochs,
            )
            self.ignite_trainer.add_event_handler(Events.EPOCH_STARTED, scheduler)

        self.model = self.model.to(self.device)
        self._metric_obj = ALL_METRICS[self.metric]
        ProgressBar(
            bar_format=None,
        ).attach(self.ignite_trainer, output_transform=lambda x: x)
        RunningAverage(output_transform=lambda x: x["loss"]).attach(self.ignite_trainer, "loss_avg")
        ProgressBar(bar_format=None).attach(self.ignite_evaluator)

    def train_step(self, engine: Engine, batch: Tuple) -> Dict[str, Tensor]:
        self.model.train()
        with torch.enable_grad():
            self.optimizer.zero_grad()
            loss = self.model(*self.prepare_batch_function(batch, self.device), return_loss=True)
            loss.backward()

            if (engine.state.iteration - 1) % self.gradient_accumulation_steps == 0:
                self.optimizer.step()
                self.optimizer.zero_grad()

            return {"loss": loss.item(), "lr": self.optimizer.param_groups[0]["lr"]}

    def validation_step(self, engine: Engine, batch: Tuple) -> Tuple[Tensor, Tensor]:
        self.model.eval()
        with torch.inference_mode():
            y_pred, y = self.model(*self.prepare_batch_function(batch, self.device), return_loss=False)
            return y_pred, y

    def run_inference(self, dl_eval):
        local_evaluator = self.ignite_evaluator
        eval_metric = self._metric_obj.metric(**self.metric_args)
        eval_metric.attach(local_evaluator, self.metric)
        local_evaluator.run(dl_eval)

        dl_eval_size = local_evaluator.state.iteration
        return local_evaluator.state.metrics[self.metric], dl_eval_size

    def run(self, dl_train, dl_dev):
        metrics = {"loss": Loss(self.model.criterion), self.metric: self._metric_obj.metric(**self.metric_args)}
        for name, metric in metrics.items():
            metric.attach(self.ignite_evaluator, name)

        @self.ignite_trainer.on(Events.EPOCH_COMPLETED)
        def log_validation_results(trainer):
            self.ignite_evaluator.run(dl_dev)
            metrics = self.ignite_evaluator.state.metrics
            logger.info(
                f"Epoch: {trainer.state.epoch} {self.metric}: {metrics[self.metric]:.3f}  Avg loss: {metrics['loss']:.5f}"
            )

        from ignite.handlers import Checkpoint, ModelCheckpoint

        checkpoint_handler = ModelCheckpoint(
            dirname=self.ckpt_dir,
            filename_pattern=self.ckpt_name,
            n_saved=1,
            create_dir=True,
            require_empty=False,
            score_name=self.metric,
            score_function=Checkpoint.get_default_score_fn(self.metric, self._metric_obj.score),
            global_step_transform=global_step_from_engine(self.ignite_trainer),
        )
        with self.ignite_evaluator.add_event_handler(
            Events.EPOCH_COMPLETED, checkpoint_handler, dict(model=self.model)
        ):
            logger.info("Trainer Run.")
            self.ignite_trainer.run(dl_train, self.max_epochs)

    def load_state_dict(self, state_dict):
        self.model.load_state_dict(state_dict)
        return self


class KNN(torch.nn.Module):
    def __init__(
        self,
        train_features: torch.Tensor,
        train_labels: torch.LongTensor,
        nb_knn: int = 10,
        device="cpu",
        T: float = 0.07,
        num_classes=50,
    ):
        super().__init__()
        self.device = device
        self.train_features = normalize(train_features.T.to(self.device), dim=1, p=2)
        self.train_labels = train_labels.view(1, -1).to(self.device)

        self.nb_knn = nb_knn
        self.T = T
        self.num_classes = num_classes

    def _get_knn_sims_and_labels(self, similarity, train_labels):
        topk_sims, indices = similarity.topk(self.nb_knn, largest=True, sorted=True)
        neighbors_labels = torch.gather(train_labels, 1, indices)
        return topk_sims, neighbors_labels

    def compute_neighbors(self, test_data: torch.Tensor) -> tuple[Tensor, Tensor]:

        similarity_rank = test_data @ self.train_features
        candidate_labels = self.train_labels.expand(len(similarity_rank), -1)
        return self._get_knn_sims_and_labels(similarity_rank, candidate_labels)

    def forward(self, test_features: torch.Tensor):
        test_features = normalize(test_features, dim=1, p=2)
        topk_sims, neighbors_labels = self.compute_neighbors(test_features)
        b = test_features.shape[0]
        topk_sims_transform = torch.nn.functional.softmax(topk_sims / self.T, 1)
        scores = torch.nn.functional.one_hot(neighbors_labels, num_classes=self.num_classes) * topk_sims_transform.view(
            b, -1, 1
        )
        return torch.sum(scores[:, : self.nb_knn, :], 1)


@dataclass
class KNNTrainer:
    num_classes: int
    device: torch.device | Literal["cpu"] = field(default_factory=lambda: torch.device("cpu"))
    nb_knn: int = 10
    temperature: float = 0.07
    metric: METRICS_TYPE = "accuracy"

    def __post_init__(self):
        self._metric_obj = ALL_METRICS[self.metric]
        self.trainer = Engine(lambda engine, batch: prepare_clip_task_batch(batch))
        ProgressBar(bar_format=None).attach(self.trainer)
        EpochOutputStore().attach(self.trainer, "output")

    def train_and_eval(self, dl_train, dl_eval) -> Tuple[float, int]:
        logger.info("KNN Feature extraction run.")
        self.trainer.run(dl_train)  # Store all features in memory, should be for most cases okay
        train_data, train_labels = zip(*self.trainer.state.output)
        train_data, train_labels = torch.cat(train_data, 0), torch.cat(train_labels, 0).long()
        knn_model = KNN(train_data, train_labels, nb_knn=self.nb_knn, num_classes=self.num_classes, T=self.temperature)

        def test_step(engine, batch):
            x, y = prepare_clip_task_batch(batch)
            return knn_model(x), y

        eval_engine = Engine(test_step)

        metrics = {self.metric: self._metric_obj.metric()}
        for name, metric in metrics.items():
            metric.attach(eval_engine, name)

        eval_engine.run(dl_eval)
        dl_eval_size = sum(1 for _ in dl_eval)
        metrics = eval_engine.state.metrics
        logger.info(f"KNN {self.metric}: {metrics[self.metric]:.3f}")
        return metrics[self.metric], dl_eval_size


if __name__ == "__main__":
    train_x = torch.randn(1000, 64)
    targets = torch.empty(1000).random_(10).long()
    test_x = torch.randn(100, 64)
    print(pad_or_trim(train_x, 65).shape)

    # model = KNN(train_x, targets, nb_knn=[10, 20, 50], num_classes=10)

    # y = model(test_x)
    # for k, v in y.items():
    # print(k, v.shape)



from xares.task import TaskConfig


def ravdess_config(encoder) -> TaskConfig:
    class_label_maps = {
        "neutral": 0,
        "calm": 1,
        "happy": 2,
        "sad": 3,
        "angry": 4,
        "fearful": 5,
        "disgust": 6,
        "surprised": 7,
    }

    config = TaskConfig(
        encoder=encoder,
        k_fold_splits=list(range(0, 4)),
        label_processor=lambda x: class_label_maps[x["emotion"]],
        name="ravdess",
        output_dim=len(class_label_maps),
        zenodo_id="14722524",
    )

    config.audio_tar_name_of_split = {fold: f"ravdess_fold_{fold}_0000000.tar" for fold in config.k_fold_splits}
    config.encoded_tar_name_of_split = {fold: f"ravdess-wds-encoded-fold-{fold}-*.tar" for fold in config.k_fold_splits}

    return config



from pathlib import Path
from typing import List

from loguru import logger
from tqdm import tqdm


def mkdir_if_not_exists(dir: Path, main_process: bool = True):
    dir = Path(dir)
    if not dir.exists() and main_process:
        logger.info(f"Creating directory {dir} ...")
        dir.mkdir(parents=True, exist_ok=True)


def download_file(url, target_path, chunk_size=8192):
    import requests

    target_path = Path(target_path) if isinstance(target_path, str) else target_path
    existing_size = target_path.stat().st_size if target_path.exists() else 0
    headers = {"Range": f"bytes={existing_size}-"}

    with requests.get(url, headers=headers, stream=True) as r:
        total_size = int(r.headers.get("content-length", 0)) + existing_size
        with (
            open(target_path, "ab") as f,
            tqdm(
                total=total_size,
                unit="iB",
                unit_scale=True,
                desc=f"Downloading {target_path.name}",
                leave=True,
                initial=existing_size,
            ) as pbar,
        ):
            for chunk in r.iter_content(chunk_size):
                size = f.write(chunk)
                pbar.update(size)

    pbar.set_description(f"Downloaded {target_path.name} ({total_size} bytes)")
    print()  # Add a newline after tqdm


def unzip_file(zip_file, dest_dir):
    import zipfile

    mkdir_if_not_exists(dest_dir)
    logger.info(f"Unzipping {zip_file} to {dest_dir} ...")
    with zipfile.ZipFile(zip_file, "r") as zip_ref:
        zip_ref.extractall(dest_dir)
    logger.info(f"Unzipping completed: {zip_file} extracted to {dest_dir}.")


def untar_file(tar_file, dest_dir):
    import tarfile

    mkdir_if_not_exists(dest_dir)
    logger.info(f"Extracting  {tar_file} to {dest_dir} ...")
    with tarfile.open(tar_file, "r|gz") as tar_ref:
        tar_ref.extractall(dest_dir)
    logger.info(f"Extracting completed: {tar_file} extracted to {dest_dir}.")


def download_zenodo_record(zenodo_id: str, target_dir: str, force_download: bool = False, temp_dir: None | str = None):
    import shutil
    import tempfile

    target_zip_path = Path(target_dir) / f"{zenodo_id}.zip"
    if not force_download and target_zip_path.exists():
        logger.info(f"{target_zip_path} already exists, skipping download.")
    else:
        temp_zip_path = Path(tempfile.gettempdir()) / f"{zenodo_id}.zip"
        zenodo_archive_url = f"https://zenodo.org/api/records/{zenodo_id}/files-archive"
        download_file(zenodo_archive_url, temp_zip_path)
        shutil.move(temp_zip_path, target_zip_path)
        logger.info(f"Downloading completed: {zenodo_id} saved to {target_zip_path}.")

    try:
        unzipped_flag = Path(target_dir) / f".unzipped"
        if not unzipped_flag.exists():
            unzip_file(target_zip_path, target_dir)
            unzipped_flag.touch()
        else:
            logger.info(f"{target_zip_path} already unzipped, skipping unzip.")
    except Exception as e:
        logger.error(f"Failed to unzip {target_zip_path} to {target_dir}: {e}.")
        logger.error(f"Remove {target_zip_path} and retry, or download manually using `download_data.sh`.")
        raise e


def attr_from_py_path(path: str, endswith: str | None = None) -> type:
    from importlib import import_module

    module_name = path.replace("/", ".").strip(".py")

    try:
        module = import_module(module_name)
    except ModuleNotFoundError:
        raise ValueError(f"Module not found: {module_name}")

    attr_list = [m for m in dir(module) if not endswith or m.endswith(endswith)]
    if len(attr_list) != 1:
        raise ValueError(f"Expected 1 class with endswith={endswith}, got {len(attr_list)}")

    return getattr(module, attr_list[0])


def download_hf_model_to_local(model_names: str | List[str], output_root: str = "."):
    # Download the model to the local directory to avoid issues under multi-processes
    if isinstance(model_names, str):
        model_names = [model_names]

    if "bert-base-uncased:tokenizer" in model_names:
        model_name = "google-bert/bert-base-uncased"
        output_path = Path(output_root) / model_name
        if not output_path.exists():
            from transformers import AutoTokenizer

            tokenizer = AutoTokenizer.from_pretrained(model_name)
            tokenizer.save_pretrained(output_path)
        model_names.remove("bert-base-uncased:tokenizer")

    if "qwen2" in model_names:
        model_name = "Qwen/Qwen2.5-0.5B"
        output_path = Path(output_root) / model_name
        if not output_path.exists():
            from transformers import AutoModelForCausalLM, AutoTokenizer

            for m in [AutoModelForCausalLM.from_pretrained(model_name), AutoTokenizer.from_pretrained(model_name)]:
                m.save_pretrained(output_path)
        model_names.remove("qwen2")

    if len(model_names) > 0:
        logger.warning(f"Models {model_names} are not supported for download")
